# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List
import torch
from meta.models.EcDc.components import CausalCNNEncoder
from lightkit.nn import Configurable
from dataclasses import dataclass, field
from meta.models.registry import register_model

from meta.data.batch import SeriesBatch
from meta.models.model import MetaModel
from .components import (
    FeatureExtractor,
    SupportSetEncoder,
    QueryEncoder,
    SupportSetQueryAttention,
    Decoder,
    TcnQueryEncoder,
    IdentityFeatureExtractor,
    LSTMSupportSetEncoder,
    LSTMQueryEncoder,
    FeedForwardQuantileDecoder,
    MultiHeadSupportSetQueryAttention,
    CNNSupportSetEncoder,
    TcnSupportSetEncoder,
)


class EncoderDecoderMetaModel(MetaModel):
    """
    Base class for meta models with a specific encoder - decoder architecture.

    Parameters
    ----------
    feature_extractor: A feature extractor that is applied to queries and support series alike
        and computes features for each time step.
    supps_enc: A support set encoder that encodes each time step of each support series.
    query_enc: A query encoder that encodes each query into a global presentation (not each time step).
    attention: Attention over encoded query and support series.
    decoder: Forecasts from attention(query, supps) and encoded query.
    """

    def __init__(
        self,
        feature_extractor: FeatureExtractor,
        supps_enc: SupportSetEncoder,
        query_enc: QueryEncoder,
        attention: SupportSetQueryAttention,
        decoder: Decoder,
    ):
        MetaModel.__init__(self)
        self.feature_extractor = feature_extractor
        self.supps_enc = supps_enc
        self.query_enc = query_enc
        self.attention = attention
        self.decoder = decoder

    def forward(
        self,
        supps: SeriesBatch,
        query: SeriesBatch,
        return_attention: bool = False,
    ) -> torch.Tensor:
        """
        Computes the forecasts from the provided queries and support set.

        Parameters
        ----------
        supps: The support set.
        query: The query.
        return_attention: If true, also return attention.
        """
        supps = self.feature_extractor(supps)
        query = self.feature_extractor(query)

        supps = self.supps_enc(supps)
        query = self.query_enc(query)

        value, attention = self.attention(query=query, supps=supps)
        result = self.decoder(query=query, value=value)
        if return_attention:
            return result, attention
        else:
            return result


@dataclass
class IwataKumagaiEcDcConfig:
    """
    Configuration class for a LSTMEncoderFeedforwardDecoder.

    See also:
        :class:`LSTMEncoderFeedforwardDecoder`
    """

    # The length of the prediction horizon
    prediction_length: int = 1
    # The quantiles the model predicts
    quantiles: List[str] = field(
        default_factory=lambda: [
            "0.02",
            "0.1",
            "0.25",
            "0.5",
            "0.75",
            "0.9",
            "0.98",
        ]
    )
    query_out_channels: int = 32
    lstm_num_layers: int = 1
    supps_out_channels: int = 32
    supps_bidirectional: bool = True
    num_heads: int = 1
    decoder_hidden_size: int = 32


@register_model
class IwataKumagaiEcDc(
    EncoderDecoderMetaModel, Configurable[IwataKumagaiEcDcConfig]
):
    """
    The base model from the paper https://arxiv.org/abs/2009.14379 by Iwata and
    Kumagai.

    Differences are quantile prediction and (optionally) longer prediction
    window.
    """

    def __init__(self, config: IwataKumagaiEcDcConfig):
        Configurable.__init__(self, config)
        EncoderDecoderMetaModel.__init__(
            self,
            feature_extractor=IdentityFeatureExtractor(),
            supps_enc=LSTMSupportSetEncoder(
                input_size=1,
                hidden_size=config.supps_out_channels,
                num_layers=config.lstm_num_layers,
                bidirectional=config.supps_bidirectional,
            ),
            query_enc=LSTMQueryEncoder(
                input_size=1,
                num_layers=config.lstm_num_layers,
                hidden_size=config.query_out_channels,
            ),
            attention=MultiHeadSupportSetQueryAttention(
                supps_size=config.supps_out_channels
                * (1 + config.supps_bidirectional),
                q_size=config.query_out_channels,
                num_heads=config.num_heads,
            ),
            decoder=FeedForwardQuantileDecoder(
                embed_size=config.query_out_channels,
                q_size=config.query_out_channels,
                hidden_size=config.decoder_hidden_size,
                prediction_length=config.prediction_length,
                num_quantiles=len(config.quantiles),
            ),
        )

    @classmethod
    def name(cls) -> str:
        return "iwata"


@dataclass
class CNNLSTMEcDcConfig:
    """
    Configuration class for a LSTMEncoderFeedforwardDecoder.

    See also:
        :class:`LSTMEncoderFeedforwardDecoder`
    """

    # The length of the prediction horizon
    prediction_length: int = 1
    # The quantiles the model predicts
    quantiles: List[str] = field(
        default_factory=lambda: [
            "0.02",
            "0.1",
            "0.25",
            "0.5",
            "0.75",
            "0.9",
            "0.98",
        ]
    )
    supps_out_channels: int = 64
    num_heads: int = 1
    attention_embed_size: int = 32


@register_model
class CNNLSTMEcDc(EncoderDecoderMetaModel, Configurable[CNNLSTMEcDcConfig]):
    """
    Support set is encoded by a CNN, query by a LSTM.
    """

    def __init__(self, config: CNNLSTMEcDcConfig):
        Configurable.__init__(self, config)
        EncoderDecoderMetaModel.__init__(
            self,
            feature_extractor=IdentityFeatureExtractor(),
            supps_enc=CNNSupportSetEncoder(
                input_size=1, out_channels=config.supps_out_channels
            ),
            query_enc=LSTMQueryEncoder(
                input_size=1,
                num_layers=1,
                hidden_size=32,
            ),
            attention=MultiHeadSupportSetQueryAttention(
                supps_size=config.supps_out_channels,
                q_size=32,
                num_heads=config.num_heads,
            ),
            decoder=FeedForwardQuantileDecoder(
                embed_size=config.attention_embed_size,
                q_size=32,
                prediction_length=config.prediction_length,
                num_quantiles=len(config.quantiles),
            ),
        )

    @classmethod
    def name(cls) -> str:
        return "cnn_iwata"


@dataclass
class TcnEcDcConfig:
    """
    Configuration class for a  TcnEcDc.

    See also:
        :class:` TcnEcDc`
    """

    # -------- decoder config ----------
    prediction_length: int = 1
    quantiles: List[str] = field(
        default_factory=lambda: [
            "0.02",
            "0.1",
            "0.25",
            "0.5",
            "0.75",
            "0.9",
            "0.98",
        ]
    )
    decoder_hidden_size: int = 64

    # -------- encoder config ----------
    num_channels: int = 64
    kernel_size: int = 2
    num_layers: int = 5

    # -------- attention config ----------
    num_heads: int = 1


@register_model
class TcnEcDc(EncoderDecoderMetaModel, Configurable[TcnEcDcConfig]):
    """
    Shared WaveNet-like encoder for query and support set.

    Multi-head attention to match the encoded query and support set.
    Feedforward decoder for multi-step quantile prediction.
    """

    def __init__(self, config: TcnEcDcConfig):
        common_encoder = CausalCNNEncoder(
            in_channels=1,
            channels=config.num_channels,
            depth=config.num_layers,
            out_channels=config.num_channels,
            kernel_size=config.kernel_size,
        )
        Configurable.__init__(self, config)
        EncoderDecoderMetaModel.__init__(
            self,
            feature_extractor=IdentityFeatureExtractor(),
            supps_enc=TcnSupportSetEncoder(encoder=common_encoder),
            query_enc=TcnQueryEncoder(
                encoder=common_encoder,
            ),
            attention=MultiHeadSupportSetQueryAttention(
                supps_size=config.num_channels,
                q_size=config.num_channels,
                num_heads=config.num_heads,
            ),
            decoder=FeedForwardQuantileDecoder(
                embed_size=config.num_channels,
                q_size=config.num_channels,
                hidden_size=config.decoder_hidden_size,
                prediction_length=config.prediction_length,
                num_quantiles=len(config.quantiles),
            ),
        )

    @classmethod
    def name(cls) -> str:
        return "tcn"



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional
from zipfile import ZipFile
from pandas.tseries.frequencies import to_offset

from src.gluonts.dataset.field_names import FieldName
from src.gluonts.dataset.repository._tsf_reader import (
    TSFReader,
    frequency_converter,
)
from src.gluonts.dataset.repository._tsf_datasets import (
    datasets,
    save_datasets,
    save_metadata,
)

from meta.datasets.gluonts import GluonTSDataModule
from meta.datasets.registry import register_data_module


@register_data_module
class DominickDataModule(GluonTSDataModule):
    """
    A data module for the dominick dataset.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _materialize(self, directory: Path) -> None:
        generate_dominick_dataset(
            dataset_path=directory / self.dataset_name,
        )

    @classmethod
    def name(cls) -> str:
        return "dm_dominick"


def generate_dominick_dataset(
    dataset_path: Path,
    prediction_length: Optional[int] = None,
):
    dataset = datasets["dominick"]
    dataset_path.mkdir(exist_ok=True)

    with TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)

        with ZipFile(dataset.download(temp_path)) as archive:
            archive.extractall(path=temp_path)

        # only one file is exptected
        reader = TSFReader(temp_path / archive.namelist()[0])
        meta, data = reader.read()

    freq = frequency_converter(meta.frequency)
    if prediction_length is None:
        if hasattr(meta, "forecast_horizon"):
            prediction_length = int(meta.forecast_horizon)
        else:
            prediction_length = default_prediction_length_from_frequency(freq)

    save_metadata(dataset_path, len(data), freq, prediction_length)

    # Impute missing start dates with unix epoch and remove time series whose
    # length is less than or equal to the prediction length
    data = [
        {**d, "start_timestamp": d.get("start_timestamp", "1970-01-01")}
        for d in data
        if len(d[FieldName.TARGET]) > prediction_length
    ]
    save_datasets(dataset_path, data, prediction_length)


def default_prediction_length_from_frequency(freq: str) -> int:
    prediction_length_map = {
        "T": 60,
        "H": 48,
        "D": 30,
        "W": 8,
        "M": 12,
        "Y": 4,
        "W-SUN": 8,  # this frequency is missing in the original gluon-ts code
    }
    try:
        freq = to_offset(freq).name
        return prediction_length_map[freq]
    except KeyError as err:
        raise ValueError(
            f"Cannot obtain default prediction length from frequency `{freq}`."
        ) from err



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Generic, List, Literal, Optional, TypeVar
import numpy as np
import numpy.typing as npt
from sklearn.preprocessing import (
    LabelEncoder,
    QuantileTransformer,
    StandardScaler,
)
from tsbench.config import Config, EnsembleConfig, ModelConfig
from tsbench.evaluations.metrics import Performance
from tsbench.evaluations.tracking import Tracker
from .transformers import PerformanceTransformer

OutputNormalization = Optional[Literal["quantile", "standard"]]

T = TypeVar("T", ModelConfig, EnsembleConfig)


class Surrogate(ABC, Generic[T]):
    """
    This class defines the interface for any surrogate model which attempts to
    predict performance metrics from model configurations.

    Subclasses may decide to only predict some performance metrics.
    """

    def __init__(
        self,
        tracker: Tracker[T],
        predict: list[str] | None = None,
        output_normalization: OutputNormalization = None,
        impute_simulatable: bool = False,
    ):
        """
        Args:
            tracker: A tracker that can be used to impute latency and number of model parameters
                into model performances. Also, it is required for some input features.
            predict: The metrics to predict. All if not provided.
            output_normalization: The type of normalization to apply to the features of each
                dataset independently. `None` applies no normalization, "quantile" applies quantile
                normalization, and "standard" transforms data to have zero mean and unit variance.
            impute_simulatable: Whether the tracker should impute latency and number of model
                parameters into the returned performance object.
        """
        self.performance_transformer = PerformanceTransformer(metrics=predict)
        self.output_normalization = output_normalization
        self.impute_simulatable = impute_simulatable
        self.tracker = tracker

    @property
    def required_cpus(self) -> int:
        """
        The number of CPUs required for fitting the surrogate.
        """
        return 1

    @property
    def required_memory(self) -> int:
        """
        The amount of memory in GiB required for fitting the surrogate.
        """
        return 1

    def fit(self, X: list[Config[T]], y: list[Performance]) -> None:
        """
        Uses the provided data to fit a model which is able to predict the
        target variables from the input.

        Args:
            X: The input configurations.
            y: The performance values associated with the input configurations.
        """
        y_numpy = self.performance_transformer.fit_transform(y)

        # If we apply any normalization, we do so independently per dataset
        if self.output_normalization is not None:
            # Initialize the transformer
            if self.output_normalization == "quantile":
                transformer = QuantileTransformer()
            else:
                transformer = StandardScaler()

            # Assign indices according to datasets
            encoder = LabelEncoder()
            dataset_indices = encoder.fit_transform(
                [x.dataset.name() for x in X]
            )

            # Then, iterate over datasets and transform the objectives
            result = np.empty_like(y_numpy)
            for i in range(len(encoder.classes_)):
                mask = dataset_indices == i
                result[mask] = transformer.fit_transform(y_numpy[mask])

            # And eventually re-assign the result
            y_numpy = result

        self._fit(X, y_numpy)

    def predict(self, X: list[Config[T]]) -> list[Performance]:
        """
        Predicts the target variables for the given inputs. Typically requires
        `fit` to be called first.

        Args:
            X: The configurations for which to predict performance metrics.

        Returns:
            The predicted performance metrics for the input configurations.
        """
        y = self._predict(X)
        performances = self.performance_transformer.inverse_transform(y)
        if not self.impute_simulatable:
            return performances

        # If the tracker is defined, latency and model parameters are set to the true values as
        # they can be simulated easily.
        for x, predicted_performance in zip(X, performances):
            true_performance = self.tracker.get_performance(x)
            # in-place operations
            predicted_performance.latency = true_performance.latency
            predicted_performance.num_model_parameters = (
                true_performance.num_model_parameters
            )
        return performances

    @abstractmethod
    def _fit(self, X: list[Config[T]], y: npt.NDArray[np.float32]) -> None:
        pass

    @abstractmethod
    def _predict(self, X: list[Config[T]]) -> npt.NDArray[np.float32]:
        pass


class DatasetFeaturesMixin:
    """
    Simple mixin which can be inherited by surrogates to signal that they
    (optionally) use dataset features.
    """



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import TYPE_CHECKING, Optional

import torch as pt
from torch import Tensor, BoolTensor
from torch.nn import functional as F


class ZScoreNormalizer(object):
    def __init__(
        self,
        eps: float = 1e-5,
        rescale_loss: bool = True,
    ) -> None:
        self._buffers = {
            "offset": None,
            "scale": None,
        }
        self.eps = eps
        self.rescale_loss = rescale_loss

    def forward_pre_hook(self, module, input):
        data = input[0]
        feats = input[1] if len(input) > 1 else None
        nan_mask = input[2] if len(input) > 2 else None
        length = input[3] if len(input) > 3 else None
        if not nan_mask.any().item():
            nan_mask = None
        if (length is None) and (nan_mask is None):
            scale = data.std(dim=1, unbiased=False, keepdim=True)
            offset = data.mean(dim=1, keepdim=True)
        else:
            if nan_mask is None:
                count = length.unsqueeze(dim=1).float()
                tensor = data
            else:
                tensor = data.masked_fill(nan_mask.unsqueeze(dim=-1), 0.0)
                if length is None:
                    count = (
                        nan_mask.logical_not().float().sum(dim=1, keepdim=True)
                    )
                else:
                    count = length.unsqueeze(
                        dim=1
                    ).float() - nan_mask.float().sum(dim=1, keepdim=True)
            offset = tensor.sum(dim=1).div(count)
            offset = offset.unsqueeze(dim=1)
            scale = tensor.pow(2).sum(dim=1).div(count)
            scale = scale.unsqueeze(dim=1)
            scale = scale.sub(offset.pow(2))
            scale = scale.sqrt()
        scale = scale + self.eps
        data = data.sub(offset).div(scale)
        self._buffers["offset"] = offset
        self._buffers["scale"] = scale

        return data, feats, nan_mask, length

    def forward_hook(self, module, input, output):
        residue, forecast, target, bc_mask, fc_mask = output
        offset = self._buffers["offset"]
        scale = self._buffers["scale"]
        rescaled_residue = residue * scale.unsqueeze(dim=2)
        rescaled_forecast = forecast * scale.unsqueeze(
            dim=2
        ) + offset.unsqueeze(dim=2)
        module.residue = rescaled_residue
        module.forecast = rescaled_forecast
        if self.rescale_loss:
            target = target * scale + offset
            residue = rescaled_residue
            forecast = rescaled_forecast
        return residue, forecast, target, bc_mask, fc_mask


class LossFunction(object):
    def __call__(self, preds: Tensor, truth: Tensor) -> Tensor:
        raise NotImplementedError

    def _aggregate_loss(
        self, loss: Tensor, loss_mask: Optional[BoolTensor]
    ) -> Tensor:
        batch_size = loss.size(0)
        if loss_mask is None:
            length = loss.size(1)
            loss = loss.view(batch_size, -1).sum(dim=1)
            if length > 0:
                loss = loss / length
        else:
            loss = loss.masked_fill(
                loss_mask.view(*loss_mask.shape, 1, 1), 0.0
            )
            loss = loss.view(batch_size, -1).sum(dim=1)
            loss_count = loss_mask.logical_not().to(pt.float).sum(dim=1)
            loss = pt.where(loss_count.eq(0), loss, loss.div(loss_count))
        return loss

    def loss_hook(self, module, input, output):
        residue, forecast, target, bc_mask, fc_mask = output
        bc_loss = self(residue, pt.zeros_like(residue))
        fc_loss = self(forecast, target.unsqueeze(dim=2).expand_as(forecast))
        if not (module.training and module.layerwise_loss):
            bc_loss = bc_loss[..., -1:, :]
            fc_loss = fc_loss[..., -1:, :]
        bc_loss = self._aggregate_loss(bc_loss, bc_mask)
        fc_loss = self._aggregate_loss(fc_loss, fc_mask)

        module.bc_loss = bc_loss.detach()
        module.fc_loss = fc_loss.detach()
        module.denominator = self._aggregate_loss(
            target.unsqueeze(dim=2), fc_mask
        )

        loss = bc_loss + module.tradeoff * fc_loss
        return loss


class MSELoss(LossFunction):
    def __call__(self, preds: Tensor, truth: Tensor) -> Tensor:
        return F.mse_loss(preds, truth, reduction="none")

    def loss_hook(self, module, input, output):
        loss = super(MSELoss, self).loss_hook(module, input, output)
        module.bc_loss = module.bc_loss.sqrt()
        module.fc_loss = module.fc_loss.sqrt()
        return loss


class MAELoss(LossFunction):
    def __call__(self, preds: Tensor, truth: Tensor) -> Tensor:
        return pt.abs(preds - truth)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from abc import ABC, abstractclassmethod
import warnings
from typing import Callable, Dict, Optional, Tuple

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import (
    Distribution,
    Beta,
    NegativeBinomial,
    StudentT,
    Normal,
    Categorical,
    MixtureSameFamily,
    Independent,
    LowRankMultivariateNormal,
    MultivariateNormal,
    TransformedDistribution,
    AffineTransform,
    Poisson,
)

from pts.distributions import (
    ZeroInflatedPoisson,
    ZeroInflatedNegativeBinomial,
    PiecewiseLinear,
    TransformedPiecewiseLinear,
    ImplicitQuantile,
    TransformedImplicitQuantile,
)
from src.gluonts.core.component import validated
from src.gluonts.torch.distributions.distribution_output import DistributionOutput
from src.gluonts.torch.modules.lambda_layer import LambdaLayer
from src.gluonts.torch.distributions.output import PtArgProj

from pts.modules.iqn_modules import ImplicitQuantileModule


class IndependentDistributionOutput(DistributionOutput):
    @validated()
    def __init__(self, dim: Optional[int] = None) -> None:
        self.dim = dim

    @property
    def event_shape(self) -> Tuple:
        if self.dim is None:
            return ()
        else:
            return (self.dim,)

    def independent(self, distr: Distribution) -> Distribution:
        if self.dim is None:
            return distr

        return Independent(distr, 1)

    def distribution(
        self, distr_args, scale: Optional[torch.Tensor] = None
    ) -> Distribution:
        distr = self.independent(self.distr_cls(*distr_args))
        if scale is None:
            return distr
        else:
            return TransformedDistribution(
                distr, [AffineTransform(loc=0, scale=scale)]
            )


class NormalOutput(IndependentDistributionOutput):
    args_dim: Dict[str, int] = {"loc": 1, "scale": 1}
    distr_cls: type = Normal

    def __init__(self, dim: Optional[int] = None) -> None:
        super().__init__(dim)
        if dim is not None:
            self.args_dim = {k: dim for k in self.args_dim}

    @classmethod
    def domain_map(cls, loc, scale):
        scale = F.softplus(scale)
        return loc.squeeze(-1), scale.squeeze(-1)


class IndependentNormalOutput(NormalOutput):
    @validated()
    def __init__(self, dim: int) -> None:
        super().__init__(dim)
        warnings.warn(
            "IndependentNormalOutput is deprecated. Use NormalOutput instead.",
            DeprecationWarning,
        )


class BetaOutput(IndependentDistributionOutput):
    args_dim: Dict[str, int] = {"concentration1": 1, "concentration0": 1}
    distr_cls: type = Beta

    def __init__(self, dim: Optional[int] = None) -> None:
        super().__init__(dim)
        if dim is not None:
            self.args_dim = {k: dim for k in self.args_dim}

    @classmethod
    def domain_map(cls, concentration1, concentration0):
        concentration1 = F.softplus(concentration1) + 1e-8
        concentration0 = F.softplus(concentration0) + 1e-8
        return concentration1.squeeze(-1), concentration0.squeeze(-1)


class PoissonOutput(IndependentDistributionOutput):
    args_dim: Dict[str, int] = {"rate": 1}
    distr_cls: type = Poisson

    def __init__(self, dim: Optional[int] = None) -> None:
        super().__init__(dim)
        if dim is not None:
            self.args_dim = {k: dim for k in self.args_dim}

    @classmethod
    def domain_map(cls, rate):
        rate_pos = F.softplus(rate).clone()

        return (rate_pos.squeeze(-1),)

    def distribution(
        self, distr_args, scale: Optional[torch.Tensor] = None
    ) -> Distribution:
        (rate,) = distr_args

        if scale is not None:
            rate *= scale

        return self.independent(Poisson(rate))


class ZeroInflatedPoissonOutput(IndependentDistributionOutput):
    args_dim: Dict[str, int] = {"gate": 1, "rate": 1}
    distr_cls: type = ZeroInflatedPoisson

    def __init__(self, dim: Optional[int] = None) -> None:
        super().__init__(dim)
        if dim is not None:
            self.args_dim = {k: dim for k in self.args_dim}

    @classmethod
    def domain_map(cls, gate, rate):
        gate_unit = torch.sigmoid(gate).clone()
        rate_pos = F.softplus(rate).clone()

        return gate_unit.squeeze(-1), rate_pos.squeeze(-1)

    def distribution(
        self, distr_args, scale: Optional[torch.Tensor] = None
    ) -> Distribution:
        gate, rate = distr_args

        if scale is not None:
            rate *= scale

        return self.independent(ZeroInflatedPoisson(gate=gate, rate=rate))


class NegativeBinomialOutput(IndependentDistributionOutput):
    args_dim: Dict[str, int] = {"total_count": 1, "logits": 1}
    distr_cls: type = NegativeBinomial

    def __init__(self, dim: Optional[int] = None) -> None:
        super().__init__(dim)
        if dim is not None:
            self.args_dim = {k: dim for k in self.args_dim}

    @classmethod
    def domain_map(cls, total_count, logits):
        total_count = F.softplus(total_count)
        return total_count.squeeze(-1), logits.squeeze(-1)

    def distribution(
        self, distr_args, scale: Optional[torch.Tensor] = None
    ) -> Distribution:
        total_count, logits = distr_args

        if scale is not None:
            logits += scale.log()

        return self.independent(
            NegativeBinomial(total_count=total_count, logits=logits)
        )


class ZeroInflatedNegativeBinomialOutput(IndependentDistributionOutput):
    args_dim: Dict[str, int] = {"gate": 1, "total_count": 1, "logits": 1}
    distr_cls: type = ZeroInflatedNegativeBinomial

    def __init__(self, dim: Optional[int] = None) -> None:
        super().__init__(dim)
        if dim is not None:
            self.args_dim = {k: dim for k in self.args_dim}

    @classmethod
    def domain_map(cls, gate, total_count, logits):
        gate = torch.sigmoid(gate)
        total_count = F.softplus(total_count)
        return gate.squeeze(-1), total_count.squeeze(-1), logits.squeeze(-1)

    def distribution(
        self, distr_args, scale: Optional[torch.Tensor] = None
    ) -> Distribution:
        gate, total_count, logits = distr_args

        if scale is not None:
            logits += scale.log()

        return self.independent(
            ZeroInflatedNegativeBinomial(
                gate=gate, total_count=total_count, logits=logits
            )
        )


class StudentTOutput(IndependentDistributionOutput):
    args_dim: Dict[str, int] = {"df": 1, "loc": 1, "scale": 1}
    distr_cls: type = StudentT

    def __init__(self, dim: Optional[int] = None) -> None:
        super().__init__(dim)
        if dim is not None:
            self.args_dim = {k: dim for k in self.args_dim}

    @classmethod
    def domain_map(cls, df, loc, scale):
        scale = F.softplus(scale)
        df = 2.0 + F.softplus(df)
        return df.squeeze(-1), loc.squeeze(-1), scale.squeeze(-1)


class StudentTMixtureOutput(DistributionOutput):
    @validated()
    def __init__(self, components: int = 1) -> None:
        self.components = components
        self.args_dim = {
            "mix_logits": components,
            "df": components,
            "loc": components,
            "scale": components,
        }

    @classmethod
    def domain_map(cls, mix_logits, df, loc, scale):
        scale = F.softplus(scale)
        df = 2.0 + F.softplus(df)
        return (
            mix_logits.squeeze(-1),
            df.squeeze(-1),
            loc.squeeze(-1),
            scale.squeeze(-1),
        )

    def distribution(
        self, distr_args, scale: Optional[torch.Tensor] = None
    ) -> Distribution:
        mix_logits, df, loc, dist_scale = distr_args

        distr = MixtureSameFamily(
            Categorical(logits=mix_logits), StudentT(df, loc, dist_scale)
        )
        if scale is None:
            return distr
        else:
            return TransformedDistribution(
                distr, [AffineTransform(loc=0, scale=scale)]
            )

    @property
    def event_shape(self) -> Tuple:
        return ()


class PiecewiseLinearOutput(DistributionOutput):
    distr_cls: type = PiecewiseLinear

    @validated()
    def __init__(self, num_pieces: int) -> None:
        super().__init__(self)
        assert (
            isinstance(num_pieces, int) and num_pieces > 1
        ), "num_pieces should be an integer larger than 1"

        self.num_pieces = num_pieces
        self.args_dim = {
            "gamma": 1,
            "slopes": num_pieces,
            "knot_spacings": num_pieces,
        }

    @classmethod
    def domain_map(cls, gamma, slopes, knot_spacings):
        # slopes of the pieces are non-negative
        slopes_proj = F.softplus(slopes) + 1e-4

        # the spacing between the knots should be in [0, 1] and sum to 1
        knot_spacings_proj = torch.softmax(knot_spacings, dim=-1)

        return gamma.squeeze(axis=-1), slopes_proj, knot_spacings_proj

    def distribution(
        self,
        distr_args,
        scale: Optional[torch.Tensor] = None,
    ) -> PiecewiseLinear:
        if scale is None:
            return self.distr_cls(*distr_args)
        else:
            distr = self.distr_cls(*distr_args)
            return TransformedPiecewiseLinear(
                distr, [AffineTransform(loc=0, scale=scale)]
            )

    @property
    def event_shape(self) -> Tuple:
        return ()


class NormalMixtureOutput(DistributionOutput):
    @validated()
    def __init__(self, components: int = 1) -> None:
        self.components = components
        self.args_dim = {
            "mix_logits": components,
            "loc": components,
            "scale": components,
        }

    @classmethod
    def domain_map(cls, mix_logits, loc, scale):
        scale = F.softplus(scale)
        return mix_logits.squeeze(-1), loc.squeeze(-1), scale.squeeze(-1)

    def distribution(
        self, distr_args, scale: Optional[torch.Tensor] = None
    ) -> Distribution:
        mix_logits, loc, dist_scale = distr_args

        distr = MixtureSameFamily(
            Categorical(logits=mix_logits), Normal(loc, dist_scale)
        )
        if scale is None:
            return distr
        else:
            return TransformedDistribution(
                distr, [AffineTransform(loc=0, scale=scale)]
            )

    @property
    def event_shape(self) -> Tuple:
        return ()


class LowRankMultivariateNormalOutput(DistributionOutput):
    @validated()
    def __init__(
        self,
        dim: int,
        rank: int,
        sigma_init: float = 1.0,
        sigma_minimum: float = 1e-3,
    ) -> None:
        self.distr_cls = LowRankMultivariateNormal
        self.dim = dim
        self.rank = rank
        self.sigma_init = sigma_init
        self.sigma_minimum = sigma_minimum
        self.args_dim = {"loc": dim, "cov_factor": dim * rank, "cov_diag": dim}

    def domain_map(self, loc, cov_factor, cov_diag):
        diag_bias = (
            self.inv_softplus(self.sigma_init**2)
            if self.sigma_init > 0.0
            else 0.0
        )

        shape = cov_factor.shape[:-1] + (self.dim, self.rank)
        cov_factor = cov_factor.reshape(shape)
        cov_diag = F.softplus(cov_diag + diag_bias) + self.sigma_minimum**2

        return loc, cov_factor, cov_diag

    def inv_softplus(self, y):
        if y < 20.0:
            return np.log(np.exp(y) - 1.0)
        else:
            return y

    @property
    def event_shape(self) -> Tuple:
        return (self.dim,)


class MultivariateNormalOutput(DistributionOutput):
    @validated()
    def __init__(self, dim: int) -> None:
        self.args_dim = {"loc": dim, "scale_tril": dim * dim}
        self.dim = dim

    def domain_map(self, loc, scale):
        d = self.dim
        device = scale.device

        shape = scale.shape[:-1] + (d, d)
        scale = scale.reshape(shape)

        scale_diag = F.softplus(
            scale * torch.eye(d, device=device)
        ) * torch.eye(d, device=device)

        mask = torch.tril(torch.ones_like(scale), diagonal=-1)
        scale_tril = (scale * mask) + scale_diag

        return loc, scale_tril

    def distribution(
        self, distr_args, scale: Optional[torch.Tensor] = None
    ) -> Distribution:
        loc, scale_tri = distr_args
        distr = MultivariateNormal(loc=loc, scale_tril=scale_tri)

        if scale is None:
            return distr
        else:
            return TransformedDistribution(
                distr, [AffineTransform(loc=0, scale=scale)]
            )

    @property
    def event_shape(self) -> Tuple:
        return (self.dim,)


class FlowOutput(DistributionOutput):
    @validated()
    def __init__(self, flow, input_size, cond_size):
        self.args_dim = {"cond": cond_size}
        self.flow = flow
        self.dim = input_size

    @classmethod
    def domain_map(cls, cond):
        return (cond,)

    def distribution(self, distr_args, scale=None):
        (cond,) = distr_args
        if scale is not None:
            self.flow.scale = scale
        self.flow.cond = cond

        return self.flow

    @property
    def event_shape(self) -> Tuple:
        return (self.dim,)


class DiffusionOutput(DistributionOutput):
    @validated()
    def __init__(self, diffusion, input_size, cond_size):
        self.args_dim = {"cond": cond_size}
        self.diffusion = diffusion
        self.dim = input_size

    @classmethod
    def domain_map(cls, cond):
        return (cond,)

    def distribution(self, distr_args, scale=None):
        (cond,) = distr_args
        if scale is not None:
            self.diffusion.scale = scale
        self.diffusion.cond = cond

        return self.diffusion

    @property
    def event_shape(self) -> Tuple:
        return (self.dim,)


class QuantilePtArgProj(PtArgProj):
    def __init__(
        self,
        in_features: int,
        output_domain_cls: nn.Module,
        args_dim: Dict[str, int],
        domain_map: Callable[..., Tuple[torch.Tensor]],
        **kwargs,
    ):
        super().__init__(in_features, args_dim, domain_map, **kwargs)
        self.output_domain_cls = output_domain_cls
        self.proj = ImplicitQuantileModule(in_features, output_domain_cls)

    def forward(self, x: torch.Tensor):
        batch_size = x.shape[0]
        forecast_length = x.shape[1]
        device = x.device
        taus = torch.rand(size=(batch_size, forecast_length), device=device)
        self.register_buffer("taus", taus)
        self.register_buffer("nn_ouput", x.clone().detach())
        predicted_quantiles = self.proj(x, taus)
        return self.domain_map(predicted_quantiles)


class ImplicitQuantileOutput(IndependentDistributionOutput):
    distr_cls: type = ImplicitQuantile
    in_features = 1
    args_dim = {"quantile_function": 1}
    output_domain_cls: type = nn.Module
    quantile_arg_proj: type = nn.Module

    @validated()
    def __init__(self, output_domain: str) -> None:
        super().__init__()
        self.set_output_domain_map(output_domain)
        self.set_args_proj()

    @classmethod
    def set_output_domain_map(cls, output_domain):
        available_domain_map_cls = {
            "Positive": nn.Softplus,
            "Real": nn.Identity,
            "Unit": nn.Softmax,
        }
        assert (
            output_domain in available_domain_map_cls.keys()
        ), "Only the following output domains are allowed: {}".format(
            available_domain_map_cls.keys()
        )
        output_domain_cls = available_domain_map_cls[output_domain]
        cls.output_domain_cls = output_domain_cls

    @classmethod
    def set_args_proj(cls):
        cls.quantile_arg_proj = QuantilePtArgProj(
            in_features=cls.in_features,
            output_domain_cls=cls.output_domain_cls,
            args_dim=cls.args_dim,
            domain_map=LambdaLayer(cls.domain_map),
        )

    @classmethod
    def domain_map(cls, *args):
        return args

    @classmethod
    def args_proj(cls, in_features):
        if in_features != cls.in_features:
            cls.in_features = in_features
            cls.set_args_proj()
        return cls.quantile_arg_proj

    def get_args_proj(self, in_features: int, prefix: Optional[str] = None):
        return self.args_proj(in_features)

    def distribution(
        self,
        distr_args,
        scale: Optional[torch.Tensor] = None,
    ) -> ImplicitQuantile:
        args_proj = self.get_args_proj(self.in_features)
        implicit_quantile_function = args_proj.proj.eval()
        distr = self.distr_cls(
            implicit_quantile_function=implicit_quantile_function,
            taus=list(args_proj.buffers())[0],
            nn_output=list(args_proj.buffers())[1],
            predicted_quantiles=distr_args,
        )
        if scale is None:
            return distr
        else:
            return TransformedImplicitQuantile(
                distr, [AffineTransform(loc=0, scale=scale)]
            )

    @property
    def event_shape(self) -> Tuple:
        return ()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Callable, List
import numpy as np
import pandas as pd
from pandas.tseries.holiday import Holiday

from src.gluonts.time_feature.holiday import indicator, distance_to_holiday


class CustomDateFeatureSet:
    """
    Implements calculation of date features. The CustomDateFeatureSet is
    applied on a pandas Series with Datetimeindex and returns a 1D array of the
    shape (1, len(date_indices)).

    Note that for lower than daily granularity the distance to the holiday is
    still computed on a per-day basis.

    Example use:

        >>> import pandas as pd
        >>> cfs = CustomDateFeatureSet([pd.to_datetime('20191129', format='%Y%m%d'),
        ...                             pd.to_datetime('20200101', format='%Y%m%d')])
        >>> date_indices = pd.date_range(
        ...     start="2019-11-24",
        ...     end="2019-12-31",
        ...     freq='D'
        ... )
        >>> cfs(date_indices)
        array([[0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
            0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
            0., 0., 0., 0., 0., 0.]])

    Example use for using a squared exponential kernel:

        >>> kernel = squared_exponential_kernel(alpha=0.5)
        >>> cfs = CustomDateFeatureSet([pd.to_datetime('20191129', format='%Y%m%d'),
        ...                             pd.to_datetime('20200101', format='%Y%m%d')], kernel)
        >>> cfs(date_indices)
        array([[3.72665317e-06, 3.35462628e-04, 1.11089965e-02, 1.35335283e-01,
            6.06530660e-01, 1.00000000e+00, 6.06530660e-01, 1.35335283e-01,
            1.11089965e-02, 3.35462628e-04, 3.72665317e-06, 1.52299797e-08,
            0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
            0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
            0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
            0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
            0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
            1.52299797e-08, 3.72665317e-06, 3.35462628e-04, 1.11089965e-02,
            1.35335283e-01, 6.06530660e-01]])
    """

    def __init__(
        self,
        reference_dates: List[pd.Timestamp],
        kernel_function: Callable[[int], int] = indicator,
    ):
        """
        Parameters
        ----------
        reference_dates
            list of panda timestamps for which features should be created.
        kernel_function
            kernel function to pass the feature value based
            on distance in days. Can be indicator function (default),
            exponential_kernel, squared_exponential_kernel or user defined.
        """
        self.reference_dates = reference_dates
        self.kernel_function = kernel_function

    def __call__(self, dates):
        """
        Transform a pandas series with timestamps to holiday features.

        Parameters
        ----------
        dates
            Pandas series with Datetimeindex timestamps.
        """
        return np.vstack(
            [
                np.hstack(
                    [
                        self.kernel_function((index - ref_date).days)
                        for index in dates
                    ]
                )
                for ref_date in self.reference_dates
            ]
        ).sum(0, keepdims=True)


class CustomHolidayFeatureSet:
    """
    Implements calculation of holiday features. The CustomHolidayFeatureSet is
    applied on a pandas Series with Datetimeindex and returns a 2D array of the
    shape (len(dates), num_features), where num_features are the number of
    holidays.

    Note that for lower than daily granularity the distance to the holiday is
    still computed on a per-day basis.

    Example use:

        >>> from pts.features import (
        ...    squared_exponential_kernel,
        ...    SpecialDateFeatureSet,
        ...    CHRISTMAS_DAY,
        ...    CHRISTMAS_EVE
        ... )
        >>> import pandas as pd
        >>> from pandas.tseries.holiday import Holiday
        >>> cfs = CustomHolidayFeatureSet([Holiday("New Years Day", month=1, day=1), Holiday("Christmas Day", month=12, day=25)])
        >>> date_indices = pd.date_range(
        ...     start="2016-12-24",
        ...     end="2016-12-31",
        ...     freq='D'
        ... )
        >>> cfs(date_indices)
        array([[1., 0., 0., 0., 0., 0., 0., 0.],
               [0., 1., 0., 0., 0., 0., 0., 0.]])

    Example use for using a squared exponential kernel:

        >>> kernel = squared_exponential_kernel(alpha=1.0)
        >>> sfs = SpecialDateFeatureSet([Holiday("New Years Day", month=1, day=1), Holiday("Christmas Day", month=12, day=25)], kernel)
        >>> sfs(date_indices)
        array([[1.00000000e+00, 3.67879441e-01, 1.83156389e-02, 1.23409804e-04,
                1.12535175e-07, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00],
               [3.67879441e-01, 1.00000000e+00, 3.67879441e-01, 1.83156389e-02,
                1.23409804e-04, 1.12535175e-07, 0.00000000e+00, 0.00000000e+00]])
    """

    def __init__(
        self,
        custom_holidays: List[Holiday],
        kernel_function: Callable[[int], int] = indicator,
    ):
        """
        Parameters
        ----------
        feature_names
            list of strings with holiday names for which features should be created.
        kernel_function
            kernel function to pass the feature value based
            on distance in days. Can be indicator function (default),
            exponential_kernel, squared_exponential_kernel or user defined.
        """
        self.custom_holidays = custom_holidays
        self.kernel_function = kernel_function

    def __call__(self, dates):
        """
        Transform a pandas series with timestamps to holiday features.

        Parameters
        ----------
        dates
            Pandas series with Datetimeindex timestamps.
        """
        return np.vstack(
            [
                np.hstack(
                    [
                        self.kernel_function(
                            distance_to_holiday(custom_holiday)(index)
                        )
                        for index in dates
                    ]
                )
                for custom_holiday in self.custom_holidays
            ]
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from src.gluonts.nursery.temporal_hierarchical_forecasting.utils.utils import (
    to_TemporalHierarchy,
)


TEMPORAL_HIERARCHIES = {
    "1min": to_TemporalHierarchy(freq_strs=["1H", "30min", "1min"]),
    "5min": to_TemporalHierarchy(freq_strs=["1H", "30min", "5min"]),
    "10min": to_TemporalHierarchy(freq_strs=["1H", "30min", "10min"]),
    "15min": to_TemporalHierarchy(freq_strs=["1H", "30min", "15min"]),
    "30min": to_TemporalHierarchy(freq_strs=["1H", "30min"]),
    "1H": to_TemporalHierarchy(freq_strs=["8H", "1H"]),
    "1D": to_TemporalHierarchy(freq_strs=["1W", "1D"]),
    "1B": to_TemporalHierarchy(freq_strs=["5D", "1D"]),
    "1W": to_TemporalHierarchy(freq_strs=["4W", "1W"]),
    "1M": to_TemporalHierarchy(freq_strs=["12M", "3M", "1M"]),
    "1Q": to_TemporalHierarchy(freq_strs=["12M", "3M"]),
}
TEMPORAL_HIERARCHIES["10T"] = TEMPORAL_HIERARCHIES["10min"]
TEMPORAL_HIERARCHIES["H"] = TEMPORAL_HIERARCHIES["1H"]
TEMPORAL_HIERARCHIES["D"] = TEMPORAL_HIERARCHIES["1D"]
TEMPORAL_HIERARCHIES["B"] = TEMPORAL_HIERARCHIES["1B"]
TEMPORAL_HIERARCHIES["M"] = TEMPORAL_HIERARCHIES["1M"]
TEMPORAL_HIERARCHIES["3M"] = TEMPORAL_HIERARCHIES["1Q"]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from __future__ import annotations
import time
import warnings
from abc import ABC, abstractmethod
from typing import Optional, Union

from torch import Tensor
from ..engine.distributed import reduce_value


class Meter(ABC):
    """
    Abstract class for meters used in metric stats. Every subclass has cached
    values for some metric.

    Parameters
    ----------
    min_mode: bool or None
        if set to true, the smaller value is considered better and vice versa;
        if set to None, no optimality is defined.
    """

    def __init__(self, min_mode: Optional[bool] = True):
        self.min_mode = min_mode
        if min_mode is True:
            self.best = float("inf")
        elif min_mode is False:
            self.best = float("-inf")
        else:
            self.best = None
        self._initialize()

    @abstractmethod
    def _initialize(self):
        pass

    def restart(self) -> None:
        """
        reset the meter to clear the cached values;

        at the same time, the optimal value so far is compared with current
        metric and updated
        """
        if self.is_optimal:
            self.best = self.value
        self._initialize()

    def update(self, *args) -> None:
        raise NotImplementedError

    @property
    @abstractmethod
    def value(self):
        """
        Read currently measured metric.
        """
        pass

    @property
    def is_optimal(self) -> bool:
        """
        Indicates whether the current metric reaches optimality.

        Raises
        ------
        TypeError
            if min_mode is None
        """
        if self.min_mode is None:
            raise TypeError("No optimal mode is defined")
        if self.min_mode is True:
            return self.value < self.best
        else:
            return self.value > self.best

    def state_dict(self) -> Dict:
        return {"best": self.best}

    def load_state_dict(self, state_dict: Dict) -> None:
        self.best = state_dict["best"]
        self._initialize()

    def __repr__(self) -> str:
        main_str = (
            f"{self.__class__.__name__}: " f"current_value = {self.value:.4f}"
        )
        if self.best is not None:
            main_str += f", best_value = {self.best:.4f}"
        return main_str


class Timer(Meter):
    """
    Timer to save elapsed time.
    """

    def __init__(self):
        super(Timer, self).__init__(None)

    @property
    def value(self) -> float:
        """
        Elapsed time in seconds since instantiation or last restart call.
        """
        return time.time() - self.start_time

    def _initialize(self) -> None:
        self.start_time = time.time()


class NumericalAverageMeter(Meter):
    """
    Maintains running average for scalar metrics.

    Call .restart() to clear added values and to start a new round of averaging
    """

    def _initialize(self) -> None:
        self._cum_values = 0.0
        self._count = 0
        self.cache = 0.0

    def update(self, value: Union[Tensor, float]) -> None:
        """
        Add the current value to be further averaged.

        Parameters
        ----------
        value : float or scalar Tensor
            the new value
        """
        if isinstance(value, Tensor):
            if value.ndim > 0:
                raise ValueError("A non-scalar tensor is provided.")
            value = value.item()
        self._cum_values += value
        self._count += 1
        self.cache = value

    @property
    def value(self) -> float:
        """
        Average of all values added since instantiation or last restart call.

        Returns
        -------
        float
            the current average
        """
        if self._count == 0:
            warnings.warn(f"Nothing have been added. Inf is returned.")
            return float("inf")
        else:
            return self._cum_values / self._count


class BatchAverageMeter(NumericalAverageMeter):
    """
    Maintains running average for a stream of batched data, in which the outer
    dimension is assumed to be batches.

    Call .restart() to clear added values and to start a new round of averaging
    """

    def __init__(
        self, min_mode: Optional[bool] = True, avg_in_batch: bool = True
    ):
        super(BatchAverageMeter, self).__init__(min_mode)
        self.averaging = avg_in_batch

    def update(self, values: Tensor) -> None:
        """
        Add a batch of values to be further averaged.

        Parameters
        ----------
        values : Tensor

        Raises
        ------
        ValueError
            if `values` is 0-dimensional
        """
        if values.ndim < 1:
            raise ValueError("A scalar tensor is provided.")
        if self.averaging:
            self._cum_values += values.mean().item() * values.size(0)
        else:
            self._cum_values += values.sum().item()
        self._count += values.size(0)
        self.cache = values.sum().item() / values.size(0)


class MeanDeviationMeter(Meter):
    """
    Maintains a stream of deviation and base values, and compute the ratio of
    their sums e.g. WAPE in demand forecasting.

    .. math::
        MD = \sigma{deviation} / \sigma{base}

    Call .restart() to clear added values and to start a new round of averaging
    """

    def __init__(self):
        super(MeanDeviationMeter, self).__init__(min_mode=True)

    def _initialize(self):
        self._cum_deviation = 0.0
        self._cum_base = 0.0
        self.cache = 0.0

    def update(self, deviation: Tensor, base: Tensor) -> None:
        """
        add new deviation and base values.

        Parameters
        ----------
        deviation : Tensor
        base : Tensor

        Raises
        ------
        ValueError
            if deviation and base are not in the same shape
        """
        if deviation.size() != base.size():
            raise ValueError(
                "`deviation` and `base` should matchh in dimensions"
            )
        self._cum_deviation += deviation.sum().item()
        self._cum_base += base.sum().item()
        self.cache = deviation.sum().div(base.sum()).item()

    @property
    def value(self) -> float:
        if self._cum_base == 0.0:
            warnings.warn("Cumulative base is 0. " "Inf is returned.")
            return float("inf")
        else:
            return self._cum_deviation / self._cum_base


class RootMeanSquareDeviationMeter(MeanDeviationMeter):
    """
    Maintains a stream of deviation and base values, and compute the following
    ratio.

    .. math::
        MD = \sqrt{\sigma{deviation^2}} / \sqrt{\sigma{base^2}}

    Call .restart() to clear added values and to start a new round of averaging
    """

    def update(self, deviation: Tensor, base: Tensor) -> None:
        """
        Add new deviation and base values.

        Parameters
        ----------
        deviation : Tensor
        base : Tensor

        Raises
        ------
        ValueError
            if deviation and base are not in the same shape
        """
        if deviation.size() != base.size():
            raise ValueError(
                "`deviation` and `base` should matchh in dimensions"
            )
        self._cum_deviation += deviation.pow(2).sum().item()
        self._cum_base += base.pow(2).sum().item()
        self.cache = (
            deviation.pow(2)
            .sum()
            .pow(0.5)
            .div(base.pow(2).sum().pow(0.5))
            .item()
        )

    @property
    def value(self) -> float:
        if self._cum_base == 0.0:
            warnings.warn("Cumulative base is 0. " "Inf is returned.")
            return float("inf")
        else:
            return self._cum_deviation**0.5 / self._cum_base**0.5



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Union, Tuple
from abc import ABC, abstractmethod
import torch
from torch import nn

from meta.data.batch import SeriesBatch


class SupportSetQueryAttention(nn.Module, ABC):
    """
    Base class for support set <--> query attention.
    """

    @abstractmethod
    def forward(
        self, query: SeriesBatch, supps: SeriesBatch
    ) -> Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor]:
        """
        Each query attends to each time point in each series of its support
        set.

        Note that we cannot use a canonical attention mechanism here because of the composition of the batches.
        The support time series are stacked along the batch dimension. Suppose a fixed support set size `x`
        batch size `bs` the batch dimension is `x * bs` i.e. a batch looks like

            [ts_11, ts_12, ...,ts_1x, ts_21, ts_22, ...ts_2x, ...,ts_bs1, ts_bs2, ..., ts_bsx].

        This is done to avoid padding and support differing support set sizes.
        However, this als means that each query should attend over multiple elements ts_*1, ...ts_*x
        along the batch dimension.

        Parameters
        ----------
        query:
            A SeriesBatch containing sequences of size [q_batch, q_size].
        supps:
             A SeriesBatch containing sequences of size [supps_batch, sequence length, supps_size].

        Returns
        -------
        torch.Tensor:
            The transformed batch containing sequences with the encoded support set series
            of size [q_batch, embedding size].
        """


class MultiHeadSupportSetQueryAttention(SupportSetQueryAttention):
    """

    Parameters
    ----------
    supps_size: The encoding dimension of the support series.
    q_size: The encoding dimension of the query series.
    embed_size: The embedding dimension for keys, queries and values.
    num_heads: The number of attention heads
    """

    def __init__(self, supps_size: int, q_size: int, num_heads: int = 1):
        super().__init__()
        assert (
            q_size % num_heads == 0
        ), "Embedding dimension must be 0 modulo number of heads."

        self.supps_embed_size = supps_size
        self.num_heads = num_heads
        self.multihead_attn = nn.MultiheadAttention(
            embed_dim=q_size,
            kdim=self.supps_embed_size,
            vdim=self.supps_embed_size,
            num_heads=num_heads,
            batch_first=True,
        )

    def forward(self, supps: SeriesBatch, query: SeriesBatch, mask=None):
        """
        Returns
        -------
        torch.Tensor:
            Aggregated values of size [batch, embed_size]
        list:
            optional, list of attention score tensors. Each tensor has size [n_supps, 1, support ts length].
            support ts length might be different from item to item in the list.
        """
        n_supps = supps.split_sections[0]
        pad_size = supps.sequences.size()[1]
        support_set = supps.sequences.reshape(
            -1, n_supps * pad_size, self.supps_embed_size
        )
        device = supps.sequences.device
        mask = torch.arange(pad_size, device=device)[None, :] < supps.lengths[
            :, None
        ].to(device)
        mask = ~mask.reshape(-1, n_supps * pad_size)

        values, attention = self.multihead_attn(
            query=query.sequences.unsqueeze(1),
            key=support_set,
            value=support_set,
            key_padding_mask=mask,
        )
        return values.squeeze(1), attention



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import argparse
import torch
import warnings

warnings.filterwarnings("ignore")
import pickle
import numpy as np
import os
from src.gluonts.dataset.repository import get_dataset
from attack_modules.attack_var import Attack
from datasets import DATASETS, get_dataset
from utils import (
    AttackResults,
    Params,
    PREDICTION_INPUT_NAMES,
    change_device,
    ts_iter,
    requires_grad_,
)
from src.gluonts.model.predictor import Predictor
from pathlib import Path
from src.gluonts.dataset.loader import InferenceDataLoader
from src.gluonts.torch.batchify import batchify

parser = argparse.ArgumentParser()

parser.add_argument("model_path", type=str, help="path to the saved model")
parser.add_argument(
    "--dataset", type=str, default="electricity", choices=DATASETS
)
parser.add_argument(
    "--attack_params", type=str, help="path to attack parameters"
)

args = parser.parse_args()


def main():
    # load predictor
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    path = Path(args.model_path)
    if "gaussian" in args.model_path and "adv" in args.model_path:
        model_type = "adv_rt"
        sparsity = args.model_path[-2]
    elif "adv" in args.model_path:
        model_type = "adv"
        sparsity = args.model_path[-2]
    elif "gaussian" in args.model_path:
        model_type = "rt"
    else:
        model_type = "vanilla"
    noise_level = 0
    for dir in args.model_path.split("/"):
        if "sigma_" in dir:
            noise_level = float(dir[6:])

    predictor = Predictor.deserialize(path)
    predictor.prediction_net = predictor.prediction_net.to(device)

    # load parameters
    prediction_length = predictor.prediction_length
    target_dim = predictor.prediction_net.target_dim
    ds = get_dataset(args.dataset, target_dim)

    test_loader = InferenceDataLoader(
        ds.test_ds,
        transform=predictor.input_transform,
        batch_size=predictor.batch_size,
        stack_fn=lambda data: batchify(data, predictor.device),
    )
    requires_grad_(predictor.prediction_net, False)
    predictor.prediction_net.eval()

    tss = list(ts_iter(ds.test_ds, ds.freq))
    true_future_targets = [
        tss[i].to_numpy()[-prediction_length:] for i in range(len(tss))
    ]
    testset_idx = 0

    params = Params(json_path=args.attack_params)
    params.device = device
    attack_params_id = args.attack_params[-6]

    attack = Attack(
        model=predictor.prediction_net,
        params=params,
        input_names=PREDICTION_INPUT_NAMES,
    )

    filename = (
        "./attack_results_var/"
        + model_type
        + "_"
        + str(prediction_length)
        + "_"
        + args.dataset
        + "_sigma_"
        + str(noise_level)
        + "_"
        + "rank"
        + str(predictor.prediction_net.distr_output.rank)
        + "_dim"
        + str(predictor.prediction_net.distr_output.dim)
    )
    if "adv" in model_type:
        filename += "_s" + str(sparsity)
    filename += "_" + attack_params_id + ".pkl"

    if not os.path.exists("./attack_results_var"):
        os.makedirs("./attack_results_var")

    with open(filename, "wb") as outp:
        for batch in test_loader:
            batch = change_device(batch, params.device)
            target = batch["past_target_cdf"]
            this_batch = target.shape[0]

            future_target = np.array(
                true_future_targets[testset_idx : testset_idx + this_batch]
            )
            best_perturbation = attack.attack_batch(
                batch,
                true_future_target=(
                    future_target
                    if device == "cpu"
                    else torch.from_numpy(future_target).float().to(device)
                ),
            )

            batch_res = AttackResults(
                batch=batch,
                perturbation=best_perturbation,
                true_future_target=future_target,
                attack_idx=params.attack_idx,
            )

            pickle.dump(batch_res, outp, pickle.HIGHEST_PROTOCOL)
            testset_idx += this_batch

            torch.cuda.empty_cache()
    print(filename)


if __name__ == "__main__":
    main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import annotations

import itertools
import math
import pickle
import random
import tempfile
from dataclasses import dataclass, field
from pathlib import Path
from typing import (
    Callable,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    TypeVar,
    Sequence,
    Tuple,
    NamedTuple,
)
from typing_extensions import Protocol, runtime_checkable
from numpy.random import RandomState
from toolz import curry

import numpy as np


@runtime_checkable
class SizedIterable(Protocol):
    def __len__(self): ...

    def __iter__(self): ...


T = TypeVar("T")

# key / value
K = TypeVar("K")
V = TypeVar("V")


def maybe_len(obj) -> Optional[int]:
    try:
        return len(obj)
    except (NotImplementedError, AttributeError):
        return None


def prod(xs):
    """
    Compute the product of the elements of an iterable object.
    """
    p = 1
    for x in xs:
        p *= x
    return p


@dataclass
class Cyclic:
    """
    Like `itertools.cycle`, but does not store the data.
    """

    iterable: SizedIterable

    def __iter__(self):
        at_least_one = False
        while True:
            for el in self.iterable:
                at_least_one = True
                yield el
            if not at_least_one:
                break

    def stream(self):
        """
        Return a continuous stream of self that has no fixed start.

        When re-iterating ``Cyclic`` it will yield elements from the start of
        the passed ``iterable``. However, this is not always desired; e.g. in
        training we want to treat training data as an infinite stream of
        values and not start at the beginning of the dataset for each epoch.

        >>> from toolz import take
        >>> c = Cyclic([1, 2, 3, 4])
        >>> assert list(take(5, c)) == [1, 2, 3, 4, 1]
        >>> assert list(take(5, c)) == [1, 2, 3, 4, 1]

        >>> s = Cyclic([1, 2, 3, 4]).stream()
        >>> assert list(take(5, s)) == [1, 2, 3, 4, 1]
        >>> assert list(take(5, s)) == [2, 3, 4, 1, 2]
        """
        return iter(self)

    def __len__(self) -> int:
        return len(self.iterable)


def batcher(iterable: Iterable[T], batch_size: int) -> Iterator[List[T]]:
    """
    Groups elements from `iterable` into batches of size `batch_size`.

    >>> list(batcher("ABCDEFG", 3))
    [['A', 'B', 'C'], ['D', 'E', 'F'], ['G']]

    Unlike the grouper proposed in the documentation of itertools, `batcher`
    doesn't fill up missing values.
    """
    it: Iterator[T] = iter(iterable)

    def get_batch():
        return list(itertools.islice(it, batch_size))

    # has an empty list so that we have a 2D array for sure
    return iter(get_batch, [])


@dataclass
class Chain:
    """
    Chain multiple iterables into a single one.

    This is a thin wrapper around ``itertools.chain``.
    """

    iterables: Collection[SizedIterable]

    def __iter__(self):
        yield from itertools.chain.from_iterable(self.iterables)

    def __len__(self) -> int:
        return sum(map(len, self.iterables))


class _SubIndex(NamedTuple):
    """
    A nested index with two layers.
    """

    item: int
    local: int


@dataclass
class Fuse:
    """
    Fuse collections together to act as single collections.

    >>> a = [0, 1, 2]
    >>> b = [3, 4, 5]
    >>> fused = Fuse([a, b])
    >>> assert len(a) + len(b) == len(fused)
    >>> list(fused[2:5])
    [2, 3, 4]
    >>> list(fused[3:])
    [3, 4, 5]

    This is similar to ``Chain``, but also allows to select directly into the
    data. While ``Chain`` creates a single ``Iterable`` out of a set of
    ``Iterable``s, ``Fuse`` creates a single ``Collection`` out of a set of
    ``Collection``s.
    """

    collections: List[Sequence]
    _lengths: List[int] = field(default_factory=list)

    def __post_init__(self):
        if not self._lengths:
            self._lengths = list(map(len, self.collections))

        self._length = sum(self._lengths)
        self._offsets = np.cumsum(self._lengths)

    def __len__(self):
        return self._length

    def _get_range(self, start: _SubIndex, stop: _SubIndex) -> "Fuse":
        first = self.collections[start.item]

        if start.item == stop.item:
            return Fuse([first[start.local : stop.local]])

        items = []

        first = first[start.local :]
        if len(first) > 0:
            items.append(first)

        for item_index in range(start.item + 1, stop.item):
            items.append(self.collections[item_index])

        items.append(self.collections[stop.item][: stop.local])
        return Fuse(items)

    def _location_for(self, idx, side="right") -> _SubIndex:
        """
        Map global index to pair of index to collection and index within that
        collection.

        >>> fuse = Fuse([[0, 0], [1, 1]])
        >>> fuse._location_for(0)
        _SubIndex(item=0, local=0)
        >>> fuse._location_for(1)
        _SubIndex(item=0, local=1)
        >>> fuse._location_for(2)
        _SubIndex(item=1, local=0)
        >>> fuse._location_for(3)
        _SubIndex(item=1, local=1)
        """
        if idx == 0 or not self:
            return _SubIndex(0, 0)

        # When the index is out of bounds, we fall back to the last element
        if idx >= len(self):
            return _SubIndex(
                len(self.collections) - 1,
                len(self.collections[-1]),
            )

        part_no = np.searchsorted(self._offsets, idx, side)

        if part_no == 0:
            local_idx = idx
        else:
            local_idx = idx - self._offsets[part_no - 1]

        return _SubIndex(int(part_no), int(local_idx))

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            start, stop, step = idx.indices(len(self))
            assert step is None or step == 1

            start_index = self._location_for(start)
            stop_index = self._location_for(stop)
            return self._get_range(start_index, stop_index)

        subindex = self._location_for(idx)
        return self.collections[subindex.item][subindex.local]

    def __iter__(self):
        yield from itertools.chain.from_iterable(self.collections)

    def __repr__(self):
        return f"Fuse<size={len(self)}>"


def split(xs: Sequence, indices: List[int]) -> List[Sequence]:
    """
    Split ``xs`` into subsets given ``indices``.

    >>> split("abcdef", [1, 3])
    ['a', 'bc', 'def']

    This is similar to ``numpy.split`` when passing a list of indices, but this
    version does not convert the underlying data into arrays.
    """

    return [
        xs[start:stop]
        for start, stop in zip(
            itertools.chain([None], indices),
            itertools.chain(indices, [None]),
        )
    ]


def split_into(xs: Sequence, n: int) -> Sequence:
    """
    Split ``xs`` into ``n`` parts of similar size.

    >>> split_into("abcd", 2)
    ['ab', 'cd']
    >>> split_into("abcd", 3)
    ['ab', 'c', 'd']
    """

    bucket_size, remainder = divmod(len(xs), n)

    # We need one fewer than `n`, since these become split positions.
    relative_splits = np.full(n - 1, bucket_size)
    # e.g. 10 by 3 -> 4, 3, 3
    relative_splits[:remainder] += 1

    return split(xs, np.cumsum(relative_splits))


@dataclass
class Cached:
    """
    An iterable wrapper, which caches values in a list while iterated.

    The primary use-case for this is to avoid re-computing the elements of the
    sequence, in case the inner iterable does it on demand.

    This should be used to wrap deterministic iterables, i.e. iterables where
    the data generation process is not random, and that yield the same
    elements when iterated multiple times.
    """

    iterable: SizedIterable
    provider: Iterable = field(init=False)
    consumed: list = field(default_factory=list, init=False)

    def __post_init__(self):
        # ensure we only iterate once over the iterable
        self.provider = iter(self.iterable)

    def __iter__(self):
        # Yield already provided values first
        yield from self.consumed

        # Now yield remaining elements.
        for element in self.provider:
            self.consumed.append(element)
            yield element

    def __len__(self) -> int:
        return len(self.iterable)


@dataclass
class PickleCached:
    """
    A caching wrapper for ``iterable`` using ``pickle`` to store cached values
    on disk.

    See :class:`Cached` for more information.
    """

    iterable: SizedIterable
    cached: bool = False
    _path: Path = field(
        default_factory=lambda: Path(
            tempfile.NamedTemporaryFile(delete=False).name
        )
    )

    def __iter__(self):
        if not self.cached:
            with open(self._path, "wb") as tmpfile:
                for batch in batcher(self.iterable, 16):
                    pickle.dump(batch, tmpfile)
                    yield from batch
            self.cached = True
        else:
            with open(self._path, "rb") as tmpfile:
                while True:
                    try:
                        yield from pickle.load(tmpfile)
                    except EOFError:
                        return

    def __del__(self):
        self._path.unlink()


@dataclass
class PseudoShuffled:
    """
    Yield items from a given iterable in a pseudo-shuffled order.
    """

    iterable: SizedIterable
    shuffle_buffer_length: int

    def __iter__(self):
        shuffle_buffer = []

        for element in self.iterable:
            shuffle_buffer.append(element)
            if len(shuffle_buffer) >= self.shuffle_buffer_length:
                yield shuffle_buffer.pop(random.randrange(len(shuffle_buffer)))

        while shuffle_buffer:
            yield shuffle_buffer.pop(random.randrange(len(shuffle_buffer)))

    def __len__(self) -> int:
        return len(self.iterable)


# can't make this a dataclass because of pytorch-lightning assumptions
class IterableSlice:
    """
    An iterable version of `itertools.islice`, i.e. one that can be iterated
    over multiple times:

        >>> isl = IterableSlice(iter([1, 2, 3, 4, 5]), 3)
        >>> list(isl)
        [1, 2, 3]
        >>> list(isl)
        [4, 5]
        >>> list(isl)
        []

    This needs to be a class to support re-entry iteration.
    """

    def __init__(self, iterable, length):
        self.iterable = iterable
        self.length = length

    def __iter__(self):
        yield from itertools.islice(self.iterable, self.length)


class SizedIterableSlice(IterableSlice):
    """
    Same as ``IterableSlice`` but also supports `len()`:

        >>> isl = SizedIterableSlice([1, 2, 3, 4, 5], 3)
        >>> len(isl)
        3
    """

    def __len__(self):
        # NOTE: This works correctly only when self.iterable supports `len()`.
        total_len = len(self.iterable)
        return min(total_len, self.length)


@dataclass
class Map:
    fn: Callable
    iterable: SizedIterable

    def __iter__(self):
        return map(self.fn, self.iterable)

    def __len__(self):
        return len(self.iterable)


@dataclass
class StarMap:
    fn: Callable
    iterable: SizedIterable

    def __iter__(self):
        return itertools.starmap(self.fn, self.iterable)

    def __len__(self):
        return len(self.iterable)


@dataclass
class Filter:
    fn: Callable
    iterable: SizedIterable

    def __iter__(self):
        return filter(self.fn, self.iterable)


@dataclass
class RandomYield:
    """
    Given a list of Iterables `iterables`, generate samples from them.

    When `probabilities` is given, sample iteratbles according to it.
    When `probabilities` is not given, sample iterables uniformly.

    When one iterable exhausts, the sampling probabilities for it will be set to 0.

        >>> from toolz import take
        >>> a = [1, 2, 3]
        >>> b = [4, 5, 6]
        >>> it = iter(RandomYield([a, b], probabilities=[1, 0]))
        >>> list(take(5, it))
        [1, 2, 3]


        >>> a = [1, 2, 3]
        >>> b = [4, 5, 6]
        >>> it = iter(RandomYield([Cyclic(a), b], probabilities=[1, 0]))
        >>> list(take(5, it))
        [1, 2, 3, 1, 2]
    """

    iterables: List[Iterable]
    probabilities: Optional[List[float]] = None
    random_state: RandomState = field(default_factory=RandomState)

    def __post_init__(self):
        if not self.probabilities:
            self.probabilities = [1.0 / len(self.iterables)] * len(
                self.iterables
            )

    def __iter__(self):
        iterators = [iter(it) for it in self.iterables]
        probs = list(self.probabilities)

        while True:
            idx = self.random_state.choice(range(len(iterators)), p=probs)
            try:
                yield next(iterators[idx])
            except StopIteration:
                probs[idx] = 0
                if sum(probs) == 0:
                    return
                probs = [prob / sum(probs) for prob in probs]


def rows_to_columns(
    rows: Sequence[Dict[K, V]],
    wrap: Callable[[Sequence[V]], Sequence[V]] = lambda x: x,
) -> Dict[K, Sequence[V]]:
    """
    Transpose rows of dicts, to one dict containing columns.

    >>> rows_to_columns([{'a': 1, 'b': 2}, {'a': 3, 'b': 4}])
    {'a': [1, 3], 'b': [2, 4]}

    This can also be understood as stacking the values of each dict onto each
    other.
    """

    if not rows:
        return {}

    column_names = rows[0].keys()

    return {
        column_name: wrap([row[column_name] for row in rows])
        for column_name in column_names
    }


def columns_to_rows(columns: Dict[K, Sequence[V]]) -> List[Dict[K, V]]:
    """
    Transpose column-orientation to row-orientation.

    >>> columns_to_rows({'a': [1, 3], 'b': [2, 4]})
    [{'a': 1, 'b': 2}, {'a': 3, 'b': 4}]
    """

    if not columns:
        return []

    column_names = columns.keys()

    return [
        dict(zip(column_names, values)) for values in zip(*columns.values())
    ]


def roundrobin(*iterables):
    """
    `roundrobin('ABC', 'D', 'EF') --> A D E B F C`

    Taken from: https://docs.python.org/3/library/itertools.html#recipes.
    """

    # Recipe credited to George Sakkis
    num_active = len(iterables)
    nexts = itertools.cycle(iter(it).__next__ for it in iterables)
    while num_active:
        try:
            for next in nexts:
                yield next()
        except StopIteration:
            # Remove the iterator we just exhausted from the cycle.
            num_active -= 1
            nexts = itertools.cycle(itertools.islice(nexts, num_active))


def partition(
    it: Iterable[T], fn: Callable[[T], bool]
) -> Tuple[List[T], List[T]]:
    """
    Partition `it` into two lists given predicate `fn`.

    This is similar to the recipe defined in Python's `itertools` docs, however
    this method consumes the iterator directly  and returns lists instead of
    iterators.
    """

    left, right = [], []

    for val in it:
        if fn(val):
            left.append(val)
        else:
            right.append(val)

    return left, right


def select(keys, source: dict, ignore_missing: bool = False) -> dict:
    """
    Select subset of `source` dictionaries.

    >>> d = {"a": 1, "b": 2, "c": 3}
    >>> select(["a", "b"], d)
    {'a': 1, 'b': 2}
    """

    result = {}

    for key in keys:
        try:
            result[key] = source[key]
        except KeyError:
            if not ignore_missing:
                raise

    return result


def trim_nans(xs, trim="fb"):
    """
    Trim the leading and/or trailing `NaNs` from a 1-D array or sequence.

    Like ``np.trim_zeros`` but for `NaNs`.
    """

    trim = trim.lower()

    start = None
    end = None

    if "f" in trim:
        for start, val in enumerate(xs):
            if not math.isnan(val):
                break

    if "b" in trim:
        for end in range(len(xs), -1, -1):
            if not math.isnan(xs[end - 1]):
                break

    return xs[start:end]


def inverse(dct: Dict[K, V]) -> Dict[V, K]:
    """
    Inverse a dictionary; keys become values and values become keys.
    """
    return {value: key for key, value in dct.items()}


_no_default = object()


@curry
def pluck_attr(seq, name, default=_no_default):
    """
    Get attribute ``name`` from elements in ``seq``.
    """

    if default is _no_default:
        return [getattr(el, name) for el in seq]

    return [getattr(el, name, default) for el in seq]


def power_set(iterable):
    """
    Generate all possible subsets of the given iterable, as tuples.

    >>> list(power_set(["a", "b"]))
    [(), ('a',), ('b',), ('a', 'b')]

    Adapted from https://docs.python.org/3/library/itertools.html#itertools-recipes
    """
    return itertools.chain.from_iterable(
        itertools.combinations(iterable, r) for r in range(len(iterable) + 1)
    )


def join_items(left, right, how="outer", default=None):
    """
    Iterate over joined dictionary items.

    Yields triples of `key`, `left_value`, `right_value`.

    Similar to SQL join statements the join behaviour is controlled by ``how``:

    * ``outer`` (default): use keys from left and right
    * ``inner``: only use keys which appear in both left and right
    * ``strict``: like ``inner``, but throws error if keys mismatch
    * ``left``: use only keys from ``left``
    * ``right``: use only keys from ``right``

    If a key is not present in either input, ``default`` is chosen instead.
    """

    if how == "outer":
        keys = {**left, **right}
    elif how == "strict":
        assert left.keys() == right.keys()
        keys = left.keys()
    elif how == "inner":
        keys = left.keys() & right.keys()
    elif how == "left":
        keys = left.keys()
    elif how == "right":
        keys = right.keys()
    else:
        raise ValueError(f"Unknown how={how}.")

    for key in keys:
        yield key, left.get(key, default), right.get(key, default)


def replace(values: Sequence[T], idx: int, value: T) -> Sequence[T]:
    """
    Replace value at index ``idx`` with ``value``.

    Like ``setitem``, but for tuples.

    >>> replace((1, 2, 3, 4), -1, 99)
    (1, 2, 3, 99)
    """
    xs = list(values)
    xs[idx] = value

    return type(values)(xs)  # type: ignore


def chop(
    at: int,
    take: int,
) -> slice:
    """
    Create slice using an index ``at`` and amount ``take``.

    >>> x = [0, 1, 2, 3, 4]
    >>> x[chop(at=1, take=2)]
    [1, 2]
    >>>
    >>> x[chop(at=-2, take=2)]
    [3, 4]
    >>> x[chop(at=3, take=-2)]
    [1, 2]
    """

    if at < 0 and take + at <= 0:
        return slice(at, None)

    if take < 0:
        return slice(at + take, at)

    return slice(at, at + take)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import annotations
import numpy as np
import pandas as pd
from dataclasses import dataclass, fields
from typing import Iterator, List, Optional
import torch
from torch.utils.data import IterableDataset, Dataset

from .dataset import TimeSeries, TimeSeriesDataset


class WeightedIndexIterator:
    """
    Iterator that caches a number of indices sampled according to given
    weights.

    This gives a great performance speedup since np.random.choice is the
    bottleneck of the data loading. This class samples and caches a certain
    number of indices and return them until new ones need to be sampled.
    """

    def __init__(self, weights: np.ndarray, num_cache: int = 1024):
        """
        Args:
            weights:  np.ndarray, containing the sample weights
            num_cache: the number of indices to cache
        """
        self.weights = weights
        self.num_cache = num_cache

    def __iter__(self):
        self.i = self.num_cache
        return self

    def __next__(self):
        if self.i >= self.num_cache:
            self.idx = np.random.choice(
                len(self.weights), p=self.weights, size=self.num_cache
            )
            self.i = 0

        idx = self.idx[self.i]
        self.i += 1
        return idx


@dataclass
class Triplet:
    """
    A triplet is composed of a support set, observed queries, and corresponding
    (unobserved) future queries.
    """

    support_set: List[
        TimeSeries
    ]  # length: support_set_size, length TimeSeries: context_length
    query_past: List[
        TimeSeries
    ]  # length: num_queries, length TimeSeries: context_length
    query_future: List[
        TimeSeries
    ]  # length: num_queries, length TimeSeries: prediction_length

    def __iter__(self):
        return (getattr(self, field.name) for field in fields(self))


class TripletDataset(Dataset[Triplet]):
    """
    The triplet dataset gets a list of queries and corresponding support set
    and returns them as triplets.
    """

    def __init__(
        self, queries: TimeSeriesDataset, support_sets: List[List[TimeSeries]]
    ):
        self.queries = queries
        self.support_sets = support_sets
        assert len(queries) == len(
            support_sets
        ), "For each query there must be exactly one support set"

    def __len__(self) -> int:
        return len(self.queries)

    def __getitem__(self, index: int) -> Triplet:
        pl = self.queries.prediction_length
        ts_l = len(self.queries[index])
        query_past = [self.queries[index][0:-pl]]
        query_future = [self.queries[index][ts_l - pl : ts_l]]
        return Triplet(self.support_sets[index], query_past, query_future)


class SamplingTripletDataset(IterableDataset[Triplet]):  # type: ignore
    """
    The sampling triplet dataset randomly samples support sets and past queries
    along with their future prediction horizon.

    All three sets consist of time series windows sliced from the original time
    series. The support set time series end before the prediction horizon
    begins to avoid time leakage. The dataset yields infinitely many items.
    Support set time series length is for now context_length.
    """

    def __init__(
        self,
        dataset: TimeSeriesDataset,
        support_set_size: int,
        num_queries: int,
        context_length: int,
        support_length: int,
        prediction_length: int,
        catch22_nn: Optional[np.ndarray] = None,
        cheat: float = 0.0,
    ):
        """
        Args:
            dataset: The dataset to sample from.
            support_set_size: The size of the support set.
            num_queries: The number of queries.
            context_length: The length of the context.
            support_length: The length of the support time series.
            prediction_length: The length of the prediction.
            catch22_nn: Contains for each index its 100 nearest neighbors w.r.t. catch22 distance.
                If not None, slices from the closest `support_set_size` time series are chosen as support set.
            cheat: If true, the query (time series to be predicted) shifted by the prediction length
                is contained in the support set, i.e. the ground truth is in the support set.
        """

        # Initialize
        super().__init__()
        self.dataset = dataset
        self.support_set_size = support_set_size
        self.num_queries = num_queries
        self.context_length = context_length
        self.support_length = support_length
        self.prediction_length = prediction_length
        self.cheat = cheat
        assert (
            not cheat or self.num_queries == 1
        ), "Cheat sampling only allows num_queries = 1"

        time_series_lengths = np.array([len(s) for s in dataset])
        time_series_weights = time_series_lengths / time_series_lengths.sum()
        self.index_iterator = iter(WeightedIndexIterator(time_series_weights))

        self.catch22_nn = catch22_nn
        assert not (
            self.catch22_nn is not None and num_queries > 1
        ), "catch22 support set selection only works with num_queries equal to one"

    def __iter__(self) -> Iterator[Triplet]:
        while True:
            (
                query_past,
                query_future,
                cheat_query,
                query_idx,
            ) = self._sample_queries()

            # We do not use the qsplit option for training
            support_set = sample_supps(
                supps_size=self.support_set_size,
                length=self.support_length,
                dataset=self.dataset,
                cheat_query=(
                    cheat_query[0] if np.random.rand() < self.cheat else None
                ),
                index_iterator=(
                    self.index_iterator
                    if self.catch22_nn is None
                    else iter(self.catch22_nn[query_idx])
                ),
            )
            yield Triplet(support_set, query_past, query_future)

    def _sample_queries(self):
        query_past = []
        query_future = []
        query_cheat = []
        for _ in range(self.num_queries):
            # First, sample a time series with probability relative to its length
            idx = next(self.index_iterator)
            series = self.dataset[idx]

            # Then, sample a slice with uniform probability
            # context should be at least prediction length long
            split_point = np.random.choice(
                np.arange(
                    self.prediction_length,
                    len(series) - self.prediction_length + 1,
                )
            )
            prediction = series[
                split_point : split_point + self.prediction_length
            ]
            context_start = max(0, split_point - self.context_length)
            context = series[context_start:split_point]

            query_past.append(context)
            query_future.append(prediction)

            # sample the start of the cheat time series
            cheat_earliest_start = max(
                0,
                context_start
                - self.support_length
                + self.context_length
                + self.prediction_length,
            )
            cheat_start = np.random.choice(
                np.arange(cheat_earliest_start, context_start + 1)
            )
            cheat_end = min(len(series), cheat_start + self.support_length)
            query_cheat.append(series[cheat_start:cheat_end])
        return query_past, query_future, query_cheat, idx


class SequentialTripletDataset(Dataset[Triplet]):  # type: ignore
    """
    The sequential triplet dataset traverses the dataset and uses the last
    prediction length slice as future query.

    The support set is sampled randomly. The length of dataset is the number of
    times series divided by the number of queries.
    """

    def __init__(
        self,
        dataset: TimeSeriesDataset,
        support_set_size: int,
        num_queries: int,
        context_length: int,
        support_length: int,
        prediction_length: int,
        support_dataset: TimeSeriesDataset = None,
        seed: Optional[int] = None,
        catch22_nn: Optional[np.ndarray] = None,
        cheat: bool = False,
    ):
        """
        Args:
            dataset: The dataset to sample from.
            support_set_size: The size of the support set.
            num_queries: The number of queries.
            context_length: The length of the context.
            support_length: The length of the support time series.
            prediction_length: The length of the prediction.
            support_dataset: The dataset to choose the support set from. If not provided `dataset` is used.
                This is used to choose the support set for the test split from the val split (technical reasons).
            seed: The random seed for sampling the support set
            catch22_nn: Contains for each index its 100 nearest neighbors w.r.t. catch22 distance.
                If not None, slices from the closest `support_set_size` time series are chosen as support set.
            cheat: If true, the query (time series to be predicted) shifted by the prediction length
                is contained in the support set, i.e. the ground truth is in the support set.
        """

        super().__init__()
        self.dataset = dataset
        self.support_set_size = support_set_size
        self.num_queries = num_queries
        self.context_length = context_length
        self.support_length = support_length
        self.prediction_length = prediction_length
        self.cheat = cheat
        assert (
            not cheat or self.num_queries == 1
        ), "Cheat sampling only allows num_queries = 1"

        self.support_dataset = support_dataset or dataset
        self.seed = seed
        if catch22_nn is None:
            time_series_lengths = np.array(
                [len(s) for s in self.support_dataset]
            )
            time_series_weights = (
                time_series_lengths / time_series_lengths.sum()
            )
            self.index_iterator = iter(
                WeightedIndexIterator(time_series_weights)
            )
        self.catch22_nn = catch22_nn
        assert not (
            self.catch22_nn is not None and num_queries > 1
        ), "catch22 support set selection only works with num_queries equal to one"

    def __len__(self) -> int:
        return len(self.dataset) // self.num_queries

    def __getitem__(self, index: int) -> Triplet:
        start = index * self.num_queries
        end = start + self.num_queries
        query_past, query_future = zip(
            *(self._last_slice(self.dataset[i]) for i in range(start, end))
        )
        # support time series should end before earliest start of future queries
        q_split = min(q.start_date for q in query_future)
        query = self.dataset[index]
        cheat_query = query[
            max(0, len(query) - self.support_length) : len(query)
        ]
        support_set = sample_supps(
            supps_size=self.support_set_size,
            length=self.support_length,
            dataset=self.support_dataset,
            q_split=q_split,
            # TODO: The seeding does not work anymore, we could use torch again in the validation split
            # but this will be slower
            # seed=self.seed + index if self.seed else None,
            cheat_query=cheat_query if np.random.rand() < self.cheat else None,
            index_iterator=(
                self.index_iterator
                if self.catch22_nn is None
                else iter(self.catch22_nn[query_past[0].item_id])
            ),
        )
        return Triplet(support_set, query_past, query_future)

    def _last_slice(self, series: TimeSeries):
        split_point = len(series) - self.dataset.prediction_length
        prediction = series[split_point : split_point + self.prediction_length]
        context = series[
            max(0, split_point - self.context_length) : split_point
        ]
        return context, prediction


def sample_supps(
    supps_size: int,
    length: int,
    dataset: TimeSeriesDataset,
    index_iterator: Iterator,
    q_split: Optional[pd.Timestamp] = None,
    # seed: Optional[int] = None,
    cheat_query: Optional[TimeSeries] = None,
):
    """
    Args:
        supps_size: The number of support time series
        length: The length of the support time series slice
        dataset: The dataset to choose support time series slices from
        index_iterator: An iterator that returns indices of the dataset sampled w.r.t. some weights.
        q_split: The latest possible end time of all support time series
        cheat_query: If not None, the cheat query is contained in the support set at a random position.
    """

    support_set = []
    for i in range(supps_size):
        series = dataset[next(index_iterator)]
        if q_split:
            # support set slice is the one closest to the prediction point of the query
            freq = series.start_date.freq
            end_point = min(
                (
                    q_split.to_period(freq) - series.start_date.to_period(freq)
                ).n,
                len(series),
            )
            if end_point <= 0:
                # this should basically never happen and is only there to not break the training in case it happens
                support_set.append(
                    TimeSeries(
                        dataset_name=series.dataset_name,
                        start_date=None,
                        values=torch.zeros(1, 1),
                        scale=torch.as_tensor([0, 1]),
                    )
                )
                continue
        else:
            # if possible choose a full-length slice
            end_point = np.random.randint(
                low=min(len(series), length), high=len(series) + 1
            )
        support_ts = series[max(0, end_point - length) : end_point]
        support_set.append(support_ts)
    if cheat_query is not None:
        support_set[np.random.choice(supps_size)] = cheat_query
    return support_set


class SuperSamplingTripletDataset(IterableDataset[Triplet]):  # type: ignore
    """
    The super sampling triplet dataset randomly samples support sets and past
    queries along with their future prediction horizon from a list of sampling
    datasets.

    First a sampling dataset is randomly chosen. Then the chosen triplet
    dataset samples support, query past and query future set. The dataset
    yields infinitely many items.
    """

    def __init__(
        self, datasets: List[SamplingTripletDataset], dataset_sampling: str
    ):
        """
        Args:
            datasets: The base datasets to sample from.
        """

        # Initialize
        super().__init__()
        datasets = list(datasets)
        self.datasets = [iter(dataset) for dataset in datasets]

        if dataset_sampling == "weighted":
            # Sample a dataset with probability relative to the total number of observations in it
            self.dataset_weights = [
                sampling_dataset.dataset.number_of_time_steps
                for sampling_dataset in datasets
            ]
            self.dataset_weights = np.array(self.dataset_weights) / sum(
                self.dataset_weights
            )
        elif dataset_sampling == "uniform":
            self.dataset_weights = None
        else:
            raise ValueError(
                f"dataset sampling {dataset_sampling} is not implemented."
            )

    def __iter__(self) -> Iterator[Triplet]:
        while True:
            index = np.random.choice(
                len(self.datasets), p=self.dataset_weights
            )
            # Return item of chosen base dataset
            yield next(self.datasets[index])



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import time
from typing import List, Optional
import torch
import torch.nn as nn
from torch.utils.tensorboard import SummaryWriter
import math
import copy
from pts.trainers.timer import Timer


class Adagrad:
    def __init__(
        self,
        task_name: str = "none",
        epochs: int = 100,
        batch_size: int = 1,
        num_batches_per_epoch: int = 50,
        num_workers: int = 1,
        pin_memory: bool = False,
        learning_rate: float = 1e-3,
        weight_decay: float = 1e-6,
        decreasing_step_size: bool = False,
        validation_freq: int = 5,
        eval_model: bool = False,
        tensorboard_path: str = "./runs/Adagrad",
        device: Optional[torch.device] = None,
    ) -> None:
        self.epochs = epochs
        self.batch_size = batch_size
        self.num_batches_per_epoch = num_batches_per_epoch
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.device = device
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.task_name = task_name
        self.decreasing_step_size = decreasing_step_size
        self.validation_freq = validation_freq
        self.eval_model = eval_model
        self.tensorboard_path = tensorboard_path

    def inference(self, model, inputs):
        output = model(*inputs)
        output = output.mean()
        if isinstance(output, (list, tuple)):
            loss = output[0]
        else:
            loss = output
        return loss

    def __call__(
        self, net: nn.Module, input_names: List[str], data_loaders
    ) -> None:
        optimizer = torch.optim.Adagrad(
            net.parameters(),
            lr=self.learning_rate,
            weight_decay=self.weight_decay,
        )

        writer = SummaryWriter(self.tensorboard_path)

        timer = Timer()

        training_iter = iter(data_loaders["training_data_loader"])
        full_batch_iter = iter(data_loaders["full_batch_loader"])

        avg_epoch_grad = 0.0
        for epoch_no in range(self.epochs):
            if self.decreasing_step_size:
                for param_group in optimizer.param_groups:
                    param_group["lr"] *= 1 / math.sqrt(epoch_no + 1)
            for batch_no in range(self.num_batches_per_epoch):
                with timer("gradient oracle"):
                    data_entry = next(training_iter)
                    optimizer.zero_grad()
                    inputs = [
                        data_entry[k].to(self.device) for k in input_names
                    ]
                    loss = self.inference(net, inputs)
                    loss.backward()
                    optimizer.step()

            # compute the gradient norm and loss over training set
            avg_epoch_loss = 0.0
            full_batch_iter = iter(data_loaders["full_batch_loader"])
            net.zero_grad()
            for i, data_entry in enumerate(full_batch_iter):
                inputs = [data_entry[k].to(self.device) for k in input_names]
                loss = self.inference(net, inputs)
                loss.backward()
                avg_epoch_loss += loss.item()
            avg_epoch_loss /= i + 1
            epoch_grad = 0.0
            for p in net.parameters():
                if p.grad is None:
                    continue
                epoch_grad += torch.norm(p.grad.data / (i + 1)).item()
            net.zero_grad()

            # compute the validation loss
            validation_loss = None
            if self.eval_model and epoch_no % self.validation_freq == 0:
                validation_iter = iter(data_loaders["validation_data_loader"])
                validation_loss = 0.0
                with torch.no_grad():
                    for i, data_entry in enumerate(validation_iter):
                        net.zero_grad()
                        inputs = [
                            data_entry[k].to(self.device) for k in input_names
                        ]
                        loss = self.inference(net, inputs)
                        validation_loss += loss.item()
                    validation_loss /= i + 1
            num_iters = (
                self.num_batches_per_epoch * (epoch_no + 1) * self.batch_size
            )
            avg_epoch_grad = (avg_epoch_grad * epoch_no + epoch_grad) / (
                epoch_no + 1
            )
            time_in_ms = timer.totals["gradient oracle"] * 1000
            writer.add_scalar(
                "gradnorm/iters",
                avg_epoch_grad,
                (epoch_no + 1) * self.num_batches_per_epoch,
            )
            writer.add_scalar("gradnorm/grads", avg_epoch_grad, num_iters)
            writer.add_scalar("gradnorm/time", avg_epoch_grad, time_in_ms)
            writer.add_scalar(
                "train_loss/iters",
                avg_epoch_loss,
                (epoch_no + 1) * self.num_batches_per_epoch,
            )
            writer.add_scalar("train_loss/grads", avg_epoch_loss, num_iters)
            writer.add_scalar("train_loss/time", avg_epoch_loss, time_in_ms)
            if self.eval_model and epoch_no % self.validation_freq == 0:
                writer.add_scalar(
                    "val_loss/iters",
                    validation_loss,
                    (epoch_no + 1) * self.num_batches_per_epoch,
                )
                writer.add_scalar("val_loss/grads", validation_loss, num_iters)
                writer.add_scalar("val_loss/time", validation_loss, time_in_ms)
                print(
                    "\nTraining Loss: {:.4f}, Test Loss: {:.4f}\n".format(
                        avg_epoch_loss, validation_loss
                    )
                )
            else:
                print(f"\nTraining Loss: {avg_epoch_loss:.4f} \n")
            print("Epoch ", epoch_no, " is done!")

        writer.close()
        print(
            "task: "
            + self.task_name
            + " on Adagrad with lr="
            + str(self.learning_rate)
            + " is done!"
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
from typing import Any


def upload_directory(
    client: Any, directory: str, bucket: str, prefix: str = ""
) -> None:
    """
    Uploads all files in the given directory (recursively) to the provided
    bucket, using the specified prefix.

    Args:
        client: The S3 client to use for uploading.
        directory: The path to the directory to upload.
        bucket: The bucket where to upload the directory.
        prefix: The prefix to use when uploading the directory.
    """
    for root, _, files in os.walk(directory):
        for file in files:
            path = os.path.join(root, file)
            with open(path, "rb") as f:
                bucket_path = os.path.relpath(path, directory)
                client.upload_fileobj(f, bucket, f"{prefix}/{bucket_path}")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Tuple

import torch
import torch.nn as nn

from src.gluonts.core.component import validated
from src.gluonts.torch.modules.distribution_output import DistributionOutput
from pts.modules import MeanScaler, NOPScaler, FeatureEmbedder


def prod(xs):
    p = 1
    for x in xs:
        p *= x
    return p


class TransformerNetwork(nn.Module):
    @validated()
    def __init__(
        self,
        input_size: int,
        d_model: int,
        num_heads: int,
        act_type: str,
        dropout_rate: float,
        dim_feedforward_scale: int,
        num_encoder_layers: int,
        num_decoder_layers: int,
        history_length: int,
        context_length: int,
        prediction_length: int,
        distr_output: DistributionOutput,
        cardinality: List[int],
        embedding_dimension: List[int],
        lags_seq: List[int],
        scaling: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.history_length = history_length
        self.context_length = context_length
        self.prediction_length = prediction_length
        self.scaling = scaling
        self.cardinality = cardinality
        self.embedding_dimension = embedding_dimension
        self.distr_output = distr_output

        assert len(set(lags_seq)) == len(
            lags_seq
        ), "no duplicated lags allowed!"
        lags_seq.sort()

        self.lags_seq = lags_seq

        self.target_shape = distr_output.event_shape

        # [B, T, input_size] -> [B, T, d_model]
        self.encoder_input = nn.Linear(input_size, d_model)
        self.decoder_input = nn.Linear(input_size, d_model)

        # [B, T, d_model] where d_model / num_heads is int
        self.transformer = nn.Transformer(
            d_model=d_model,
            nhead=num_heads,
            num_encoder_layers=num_encoder_layers,
            num_decoder_layers=num_decoder_layers,
            dim_feedforward=dim_feedforward_scale * d_model,
            dropout=dropout_rate,
            activation=act_type,
        )

        self.proj_dist_args = distr_output.get_args_proj(d_model)

        self.embedder = FeatureEmbedder(
            cardinalities=cardinality,
            embedding_dims=embedding_dimension,
        )

        if scaling:
            self.scaler = MeanScaler(keepdim=True)
        else:
            self.scaler = NOPScaler(keepdim=True)

        # mask
        self.register_buffer(
            "tgt_mask",
            self.transformer.generate_square_subsequent_mask(
                prediction_length
            ),
        )

    @staticmethod
    def get_lagged_subsequences(
        sequence: torch.Tensor,
        sequence_length: int,
        indices: List[int],
        subsequences_length: int = 1,
    ) -> torch.Tensor:
        """
        Returns lagged subsequences of a given sequence.
        Parameters
        ----------
        sequence : Tensor
            the sequence from which lagged subsequences should be extracted.
            Shape: (N, T, C).
        sequence_length : int
            length of sequence in the T (time) dimension (axis = 1).
        indices : List[int]
            list of lag indices to be used.
        subsequences_length : int
            length of the subsequences to be extracted.
        Returns
        --------
        lagged : Tensor
            a tensor of shape (N, S, C, I), where S = subsequences_length and
            I = len(indices), containing lagged subsequences. Specifically,
            lagged[i, j, :, k] = sequence[i, -indices[k]-S+j, :].
        """
        assert max(indices) + subsequences_length <= sequence_length, (
            f"lags cannot go further than history length, found lag {max(indices)} "
            f"while history length is only {sequence_length}"
        )
        assert all(lag_index >= 0 for lag_index in indices)

        lagged_values = []
        for lag_index in indices:
            begin_index = -lag_index - subsequences_length
            end_index = -lag_index if lag_index > 0 else None
            lagged_values.append(sequence[:, begin_index:end_index, ...])
        return torch.stack(lagged_values, dim=-1)

    def create_network_input(
        self,
        feat_static_cat: torch.Tensor,  # (batch_size, num_features)
        feat_static_real: torch.Tensor,
        # (batch_size, num_features, history_length)
        past_time_feat: torch.Tensor,
        past_target: torch.Tensor,  # (batch_size, history_length, 1)
        past_observed_values: torch.Tensor,  # (batch_size, history_length)
        future_time_feat: Optional[
            torch.Tensor
        ],  # (batch_size, num_features, prediction_length)
        # (batch_size, prediction_length)
        future_target: Optional[torch.Tensor],
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Creates inputs for the transformer network.

        All tensor arguments should have NTC layout.
        """

        if future_time_feat is None or future_target is None:
            time_feat = past_time_feat[
                :, self.history_length - self.context_length :, ...
            ]
            sequence = past_target
            sequence_length = self.history_length
            subsequences_length = self.context_length
        else:
            time_feat = torch.cat(
                (
                    past_time_feat[
                        :, self.history_length - self.context_length :, ...
                    ],
                    future_time_feat,
                ),
                dim=1,
            )
            sequence = torch.cat((past_target, future_target), dim=1)
            sequence_length = self.history_length + self.prediction_length
            subsequences_length = self.context_length + self.prediction_length

        # (batch_size, sub_seq_len, *target_shape, num_lags)
        lags = self.get_lagged_subsequences(
            sequence=sequence,
            sequence_length=sequence_length,
            indices=self.lags_seq,
            subsequences_length=subsequences_length,
        )

        # scale is computed on the context length last units of the past target
        # scale shape is (batch_size, 1, *target_shape)
        _, scale = self.scaler(
            past_target[:, -self.context_length :, ...],
            past_observed_values[:, -self.context_length :, ...],
        )
        embedded_cat = self.embedder(feat_static_cat)

        # in addition to embedding features, use the log scale as it can help prediction too
        # (batch_size, num_features + prod(target_shape))
        static_feat = torch.cat(
            (
                embedded_cat,
                feat_static_real,
                (
                    torch.log(scale)
                    if len(self.target_shape) == 0
                    else torch.log(scale.squeeze(1))
                ),
            ),
            dim=1,
        )

        repeated_static_feat = static_feat.unsqueeze(1).expand(
            -1, subsequences_length, -1
        )

        # (batch_size, sub_seq_len, *target_shape, num_lags)
        lags_scaled = lags / scale.unsqueeze(-1)

        # from (batch_size, sub_seq_len, *target_shape, num_lags)
        # to (batch_size, sub_seq_len, prod(target_shape) * num_lags)
        input_lags = lags_scaled.reshape(
            (
                -1,
                subsequences_length,
                len(self.lags_seq) * prod(self.target_shape),
            )
        )

        # (batch_size, sub_seq_len, input_dim)
        inputs = torch.cat(
            (input_lags, time_feat, repeated_static_feat), dim=-1
        )

        return inputs, scale, static_feat


class TransformerTrainingNetwork(TransformerNetwork):
    # noinspection PyMethodOverriding,PyPep8Naming
    def forward(
        self,
        feat_static_cat: torch.Tensor,
        feat_static_real: torch.Tensor,
        past_time_feat: torch.Tensor,
        past_target: torch.Tensor,
        past_observed_values: torch.Tensor,
        future_time_feat: torch.Tensor,
        future_target: torch.Tensor,
    ) -> torch.Tensor:
        """
        Computes the loss for training Transformer, all inputs tensors representing time series have NTC layout.
        Parameters
        ----------
        feat_static_cat : (batch_size, num_features)
        feat_static_real: torch.Tensor,  # (batch_size, num_features)
        past_time_feat : (batch_size, history_length, num_features)
        past_target : (batch_size, history_length, *target_shape)
        past_observed_values : (batch_size, history_length, *target_shape, seq_len)
        future_time_feat : (batch_size, prediction_length, num_features)
        future_target : (batch_size, prediction_length, *target_shape)
        Returns
        -------
        Loss with shape (batch_size, context + prediction_length, 1)
        """

        # create the inputs for the encoder
        inputs, scale, _ = self.create_network_input(
            feat_static_cat=feat_static_cat,
            feat_static_real=feat_static_real,
            past_time_feat=past_time_feat,
            past_target=past_target,
            past_observed_values=past_observed_values,
            future_time_feat=future_time_feat,
            future_target=future_target,
        )

        enc_input = inputs[:, : self.context_length, ...]  # F.slice_axis(
        #     inputs, axis=1, begin=0, end=self.context_length
        # )
        dec_input = inputs[:, self.context_length :, ...]  # F.slice_axis(
        #     inputs, axis=1, begin=self.context_length, end=None
        # )

        # pass through encoder [T, B, b_model]
        enc_out = self.transformer.encoder(
            self.encoder_input(enc_input).permute(1, 0, 2)
        )

        # input to decoder
        dec_output = self.transformer.decoder(
            self.decoder_input(dec_input).permute(1, 0, 2),
            enc_out,  # memory
            tgt_mask=self.tgt_mask,
        )

        # compute loss
        distr_args = self.proj_dist_args(dec_output.permute(1, 0, 2))
        distr = self.distr_output.distribution(distr_args, scale=scale)
        loss = -distr.log_prob(future_target)

        return loss.mean()


class TransformerPredictionNetwork(TransformerNetwork):
    def __init__(self, num_parallel_samples: int = 100, **kwargs) -> None:
        super().__init__(**kwargs)
        self.num_parallel_samples = num_parallel_samples

        # for decoding the lags are shifted by one,
        # at the first time-step of the decoder a lag of one corresponds to the last target value
        self.shifted_lags = [l - 1 for l in self.lags_seq]

    def sampling_decoder(
        self,
        static_feat: torch.Tensor,
        past_target: torch.Tensor,
        time_feat: torch.Tensor,
        scale: torch.Tensor,
        enc_out: torch.Tensor,
    ) -> torch.Tensor:
        """
        Computes sample paths by decoding from the transformer.
        Parameters
        ----------
        static_feat : Tensor
            static features. Shape: (batch_size, num_static_features).
        past_target : Tensor
            target history. Shape: (batch_size, history_length, 1).
        time_feat : Tensor
            time features. Shape: (batch_size, prediction_length, num_time_features).
        scale : Tensor
            tensor containing the scale of each element in the batch. Shape: (batch_size, ).
        enc_out: Tensor
            output of the encoder. Shape: (batch_size, num_cells)
        Returns
        --------
        sample_paths : Tensor
            a tensor containing sampled paths. Shape: (batch_size, num_sample_paths, prediction_length).
        """

        # blows-up the dimension of each tensor to batch_size * self.num_parallel_samples for increasing parallelism
        repeated_past_target = past_target.repeat_interleave(
            repeats=self.num_parallel_samples, dim=0
        )

        repeated_time_feat = time_feat.repeat_interleave(
            repeats=self.num_parallel_samples, dim=0
        )

        repeated_static_feat = static_feat.repeat_interleave(
            repeats=self.num_parallel_samples, dim=0
        ).unsqueeze(1)

        repeated_enc_out = enc_out.repeat_interleave(
            repeats=self.num_parallel_samples, dim=1
        )

        repeated_scale = scale.repeat_interleave(
            repeats=self.num_parallel_samples, dim=0
        )

        future_samples = []

        # for each future time-units we draw new samples for this time-unit and update the state
        for k in range(self.prediction_length):
            lags = self.get_lagged_subsequences(
                sequence=repeated_past_target,
                sequence_length=self.history_length + k,
                indices=self.shifted_lags,
                subsequences_length=1,
            )

            # (batch_size * num_samples, 1, *target_shape, num_lags)
            lags_scaled = lags / repeated_scale.unsqueeze(1)
            # lags_scaled = F.broadcast_div(
            #     lags, repeated_scale.expand_dims(axis=-1)
            # )

            # from (batch_size * num_samples, 1, *target_shape, num_lags)
            # to (batch_size * num_samples, 1, prod(target_shape) * num_lags)
            input_lags = lags_scaled.reshape(
                shape=(-1, 1, prod(self.target_shape) * len(self.lags_seq))
            )

            # (batch_size * num_samples, 1, prod(target_shape) * num_lags + num_time_features + num_static_features)
            dec_input = torch.cat(
                (
                    input_lags,
                    repeated_time_feat[:, k : k + 1, :],
                    repeated_static_feat,
                ),
                dim=-1,
            )

            dec_output = self.transformer.decoder(
                self.decoder_input(dec_input).permute(1, 0, 2),
                repeated_enc_out,
            )

            distr_args = self.proj_dist_args(dec_output.permute(1, 0, 2))

            # compute likelihood of target given the predicted parameters
            distr = self.distr_output.distribution(
                distr_args, scale=repeated_scale
            )

            # (batch_size * num_samples, 1, *target_shape)
            new_samples = distr.sample()

            # (batch_size * num_samples, seq_len, *target_shape)
            repeated_past_target = torch.cat(
                (repeated_past_target, new_samples), dim=1
            )
            future_samples.append(new_samples)

        # reset cache of the decoder
        # self.transformer.decoder.cache_reset()

        # (batch_size * num_samples, prediction_length, *target_shape)
        samples = torch.cat(future_samples, dim=1)

        # (batch_size, num_samples, *target_shape, prediction_length)
        return samples.reshape(
            (
                (-1, self.num_parallel_samples)
                + self.target_shape
                + (self.prediction_length,)
            )
        )

    def forward(
        self,
        feat_static_cat: torch.Tensor,
        feat_static_real: torch.Tensor,
        past_time_feat: torch.Tensor,
        past_target: torch.Tensor,
        past_observed_values: torch.Tensor,
        future_time_feat: torch.Tensor,
    ) -> torch.Tensor:
        """
        Predicts samples, all tensors should have NTC layout.
        Parameters
        ----------
        feat_static_cat : (batch_size, num_features)
        feat_static_real : (batch_size, num_features)
        past_time_feat : (batch_size, history_length, num_features)
        past_target : (batch_size, history_length, *target_shape)
        past_observed_values : (batch_size, history_length, *target_shape)
        future_time_feat : (batch_size, prediction_length, num_features)
        Returns predicted samples
        -------
        """

        # create the inputs for the encoder
        inputs, scale, static_feat = self.create_network_input(
            feat_static_cat=feat_static_cat,
            feat_static_real=feat_static_real,
            past_time_feat=past_time_feat,
            past_target=past_target,
            past_observed_values=past_observed_values,
            future_time_feat=None,
            future_target=None,
        )

        # pass through encoder
        enc_out = self.transformer.encoder(
            self.encoder_input(inputs).permute(1, 0, 2)
        )

        return self.sampling_decoder(
            past_target=past_target,
            time_feat=future_time_feat,
            static_feat=static_feat,
            scale=scale,
            enc_out=enc_out,
        )
